# !/usr/bin/python
# coding=utf-8
import os, sys
import re
import glob
import importlib
import inspect
from functools import partial

from PySide2 import QtCore, QtGui, QtWidgets
from PySide2.QtUiTools import QUiLoader

from pythontk import File, Str, Iter, hasAttribute, setAttributes
from uitk.styleSheet import StyleSheet


class MainWindow(QtWidgets.QMainWindow):
	def __init__(self, switchboard_instance, file):
		'''Represents a main window in a GUI application.
		Inherits from QtWidgets.QMainWindow class, providing additional functionality for 
		managing user interface (UI) elements.

		Parameters:
			switchboard_instance (obj): An instance of the switchboard class
			file_path (str): The full path to the UI file

		Attributes:
			switchboard_instance (obj): The instance of the switchboard class
			path (str): The path of the UI file
			level (int): The level of the UI in the hierarchy (0: init (root), 1: base menus, 2: sub menus, 3: parent menus, 4: popup menus)
			isSubmenu (bool): Whether the UI is a submenu
			isInitialized (bool): Whether the UI has been initialized (True after the UI is first shown)
			isConnected (bool): Whether the UI is connected
			preventHide (bool): Whether hiding the UI should be prevented
			connectOnShow (bool): Whether the UI should be connected on show
			base (str): The base of the UI name
			tags (list): The tags of the UI name (trailing strings in the UI name preceded by a hashtag used to define special behaviors)
			sizeX (int): The width of the UI
			sizeY (int): The height of the UI
			widgets (list): The list of widgets in the UI

		Properties:
			name (str): The name of the UI
			ui (obj): The current UI
			<UI>.name (str): The UI filename
			<UI>.alias (str): Get or set an alternate attribute name that can be used to access the UI
			<UI>.base (str): The base UI name
			<UI>.path (str): The directory path containing the UI file
			<UI>.tags (list): Any UI tags as a list
			<UI>.level (int): The UI level
			<UI>.sizeX (int): The original width of the UI
			<UI>.sizeY (int): The original height of the UI
			<UI>.isCurrentUi (bool): True if the UI is set as current
			<UI>.isSubmenu (bool): True if the UI is a submenu
			<UI>.isInitialized (bool): True after the UI is first shown
			<UI>.connected (bool): True if the UI is connected. If set to True, the UI will be set as current and connections established
			<UI>.isConnected (bool): True if the UI is connected to its slots
			<UI>.connectOnShow (bool): While True, the UI will be set as current on show
			<UI>.show: An override of the built-in hide method
			<UI>.preventHide (bool): While True, the hide method is disabled
			<UI>.hide: An override of the built-in hide method
			<UI>.initWidgets (function): Initialize widgets.
			<UI>.widgets (list): All the widgets of the UI
			<UI>.slots (obj): The slots class instance
		'''
		super().__init__()

		self.sb = switchboard_instance
		self.name = File.formatPath(file, 'name')
		self.path = File.formatPath(file, 'path')
		self.level = self.sb._getUiLevelFromDir(file)
		self.isSubmenu = self.level==2
		self.isInitialized = False
		self.isConnected = False
		self.preventHide = False
		self.connectOnShow = True
		self.base = next(iter(self.name.split('_')))
		self.tags = self.name.split('#')[1:]
		self.sizeX = self.frameGeometry().width()
		self.sizeY = self.frameGeometry().height()
		self._widgets = set()
		self._deferred = {}

		ui = self.sb.load(file)
		self.setWindowFlags(ui.windowFlags())
		self.setCentralWidget(ui.centralWidget())

	def __getattr__(self, attr_name):
		"""Looks for the widget in the parent class.
		If found, the widget is initialized and returned, else an AttributeError is raised.

		Parameters:
			attr_name (str): the name of the attribute being accessed.

		Return:
			() The value of the widget attribute if it exists, or raises an AttributeError
			if the attribute cannot be found.
  
		Raises:
			AttributeError: if the attribute does not exist in the current instance
			or the parent class.
		"""
		found_widget = self.sb._getWidgetFromUi(self, attr_name)
		if found_widget:
			self.sb.initWidgets(self, found_widget)
			return found_widget
		raise AttributeError(f'{self.__class__.__name__} has no attribute `{attr_name}`')

	def event(self, event):
		if event.type() == QtCore.QEvent.ChildPolished:
			child = event.child()
			self.on_child_polished(child)
		return super().event(event)

	def defer(self, func, *args, priority=0):
		"""Defer execution of a function until later. The function is added to a dictionary of deferred 
		methods, with a specified priority. Lower priority values will be executed before higher ones.
		
		Parameters:
			func (function): The function to defer.
			*args: Any arguments to be passed to the function.
			priority (int, optional): The priority of the deferred method. Lower values will be executed 
					first. Defaults to 0.
		"""
		method = partial(func, *args)
		if priority in self._deferred:
			self._deferred[priority] += (method,)
		else:
			self._deferred[priority] = (method,)

	def trigger_deferred(self):
		"""Executes all deferred methods, in priority order. Any arguments passed to the deferred functions
		will be applied at this point. Once all deferred methods have executed, the dictionary is cleared.
		"""
		for priority in sorted(self._deferred):
			for method in self._deferred[priority]:
				method()
		self._deferred.clear()

	@property
	def name(self):
		return self.objectName()

	@name.setter
	def name(self, new_name):
		try:
			delattr(self.sb, self.objectName())
		except AttributeError:
			pass
		self.setObjectName(new_name)
		setattr(self.sb, new_name, self)

	@property
	def widgets(self):
		return list(self._widgets) or self.sb.initWidgets(self, self.findChildren(QtWidgets.QWidget), returnAllWidgets=True)

	@property
	def slots(self):
		return self.sb.getSlots(self)

	@property
	def isCurrentUi(self):
		return self==self.sb.getCurrentUi()

	def setAsCurrent(self):
		self.sb.setCurrentUi(self)

	def connect(self):
		self.sb.setConnections(self)

	def on_child_polished(self, w): #called after child polished event.
		if w not in self._widgets:
			self.sb.initWidgets(self, w)
			# print ('on_child_polished:', w.ui.name.ljust(30), w.name.ljust(15), id(w)) #debug
			self.trigger_deferred()
			if self.isConnected:
				self.sb.connectSlots(self, w)

	def setVisible(self, state): #called every time the after widget is shown or hidden on screen.
		if state: #visible
			if self.connectOnShow and not self.isConnected:
				self.sb.setConnections(self)
			self.activateWindow()
			self.isInitialized = True
		elif self.preventHide: #invisible
			return
		super().setVisible(state)



class Switchboard(QUiLoader, StyleSheet):
	'''Load dynamic UI, assign convenience properties, and handle slot connections.

	Properties:
		sb: The instance of this class holding all properties.
		sb.ui: Returns the current UI.
		sb.<uiFileName>: Accesses the UI loaded from uiFileName.
		sb.<customWidgetClassName>: Accesses the custom widget with the specified class name.

	Parameters:
		parent (obj): A QtObject derived class.
		dynui (str/obj): Set the directory of the dynamic UI, or give the dynamic UI objects.
		widgets (str/obj): Set the directory of any custom widgets, or give the widget objects.
		slots (str/obj): Set the directory of where the slot classes will be imported, or give the slot class itself.
		preloadUi (bool): Load all UI immediately. Otherwise UI will be loaded as required.
		style (str): Stylesheet color mode. ie. 'standard', 'dark', None
		submenuStyle (str): The stylesheet color mode for submenus.

	Methods:
		loadUi(uiPath): Load the UI file located at uiPath.
		loadAllUi(): Load all UI files in the UI directory.
		registerWidget(widget): Register the specified widget.
		connectSlots(slotClass, ui=None): Connect the slots in the specified slot class to the specified UI.

	Attributes:
		defaultDir: The default directory.
		defaultSignals: A dictionary of the default signals to be connected per widget type.

	Default Directories:
		The default directory is the calling module's directory.
		If any of the given file paths are not a full path, they will be treated as relative to the currently set path.

	Example:
		1. Create a subclass of Switchboard and define the slots for the UI events.
			class MySwitchboard(Switchboard):
				def __init__(self, parent=None, **kwargs):
					super().__init__(parent)
					...
				def my_slot(self):
					...
		2. Instantiate the subclass and show the UI.
			sb = MySwitchboard()
			sb.ui.show()
		3. Run the app, show the window, wait for input, then terminate program with the status code returned from app.
			exit_code = sb.app.exec_()
			if exit_code != -1:
				sys.exit(exit_code)
	'''
	app = QtWidgets.QApplication.instance() or QtWidgets.QApplication(sys.argv) #return the existing QApplication object, or create a new one if none exists.

	def __init__(self, parent=None, dynui=None, widgets=None, slots=None, preloadUi=False, style='standard', submenuStyle='dark'):
		super().__init__(parent)
		'''
		'''
		calling_frame = inspect.currentframe().f_back
		calling_file = calling_frame.f_code.co_filename
		self.defaultDir = os.path.abspath(os.path.dirname(calling_file))
		self.moduleDir = File.getFilepath(__file__)

		self.dynui = dynui or f'{self.moduleDir}/ui' #use the relative filepath of this module if None is given.
		self.widgets = widgets or f'{self.moduleDir}/widgets'
		self.slots = slots or f'{self.moduleDir}/slots'

		self.style = style
		self.submenuStyle = submenuStyle

		self._uiHistory = [] #A list of previously loaded ui.
		self._wgtHistory = [] #A list of previously used widgets.
		self._registeredWidgets = {} #A dict of all registered custom widgets.
		self._loadedUi = set() #A set of all loaded ui.
		self._synced_pairs = set() #A set of hashed values representing widgets that have synced values.
		self._gcProtect = set() #A set of widgets to be protected from garbage collection.

		self.defaultSignals = { #the signals to be connected per widget type. Values can be a list or a single item string.
			'QAction':'triggered',
			'QLabel':'released',
			'QPushButton':'released',
			'QListWidget':'itemClicked',
			'QTreeWidget':'itemClicked',
			'QComboBox':'currentIndexChanged',
			'QSpinBox':'valueChanged',
			'QDoubleSpinBox':'valueChanged',
			'QCheckBox':'clicked',
			'QRadioButton':'released',
			'QLineEdit':['editingFinished'],
			'QTextEdit':'textChanged',
			'QSlider':'sliderMoved',
			'QProgressBar':'valueChanged',
		}

		if preloadUi:
			self.loadAllUi()

	def __getattr__(self, attr_name):
		'''If an unknown attribute matches the name of a ui in the current ui directory; load and return it.
		Else, if an unknown attribute matches the name of a custom widget in the widgets directory; register and return it.
		If no match is found raise an attribute error.

		Return:
			(obj) ui or widget.
		'''
		# Check if the attribute matches a ui file
		ui_path = File.formatPath(self.dynui, 'path')
		found_ui = next((f for f in glob.iglob(f'{ui_path}/**/{attr_name}.ui', recursive=True)), None)
		if found_ui:
			ui = self.loadUi(found_ui)
			return ui

		# Check if the attribute matches a widget file
		widget_path = File.formatPath(self.widgets, 'path')
		widget_name = Str.setCase(attr_name, 'camel')
		found_widget = next((f for f in glob.iglob(f'{widget_path}/**/{widget_name}.py', recursive=True)), None)
		if found_widget:
			widget = self.registerWidgets(found_widget)
			if isinstance(widget, list):
				widget = next(iter(w for w in widget if w.__name__.lower()==attr_name.lower()), None) #check if any of the widgets in the list has a name that matches the attribute name in a case-insensitive manner.
			if widget and widget.__name__.lower()==attr_name.lower(): #If the widget's name matches the attribute name, the widget is returned.
				return widget

		raise AttributeError(f'{self.__class__.__name__} has no attribute `{attr_name}`')


	@property
	def dynui(self) -> str:
		'''Get the directory where the dynamic ui are stored.

		Return:
			(str) directory path.
		'''
		
		if not hasAttribute(self, '_uiLoc'): #does not invoke __getattr__.
			self._uiLoc = self.defaultDir
			return self._uiLoc
		return self._uiLoc


	@dynui.setter
	def dynui(self, d) -> None:
		'''Set the directory where the dynamic ui are located.

		Parameters:
			d (str): The directory path where the dynamic ui are located.
				If the given dir is not a full path, it will be treated as relative to the default path.
		'''
		if isinstance(d, str):
			isAbsPath = os.path.isabs(d)
			self._uiLoc = d if isAbsPath else os.path.join(self.defaultDir, d) #if the given dir is not a full path, treat it as relative to the default path.
		else: #store object.
			self.setWorkingDirectory(d) #set QUiLoader working path.
			self._uiLoc = d


	@property
	def widgets(self) -> str:
		'''Get the directory where any custom widgets are stored.

		Return:
			(str) directory path.
		'''
		if not hasAttribute(self, '_widgetLoc'): #does not invoke __getattr__.
			self._widgetLoc = self.defaultDir
			return self._widgetLoc
		return self._widgetLoc


	@widgets.setter
	def widgets(self, d) -> None:
		'''Set the directory where any custom widgets are stored.

		Parameters:
			d (str): The directory path where any custom widgets are located.
				If the given dir is not a full path, it will be treated as relative to the default path.
		'''
		if isinstance(d, str):
			isAbsPath = os.path.isabs(d)
			self._widgetLoc = d if isAbsPath else os.path.join(self.defaultDir, d) #if the given dir is not a full path, treat it as relative to the default path.
		else: #store object.
			self.addPluginPath(d) #set QUiLoader working path.
			self._widgetLoc = d


	@property
	def slots(self) -> str:
		'''Get the directory where the slot classes will be imported from.

		Return:
			(str)(obj) slots class directory path or slots class object.
		'''
		try:
			return self._slotLoc

		except AttributeError as error:
			self._slotLoc = self.defaultDir
			return self._slotLoc


	@slots.setter
	def slots(self, d) -> None:
		'''Set the directory where the slot classes will be imported from.

		Parameters:
			d (str): The directory path where the slot classes are located.
				If the given dir is not a full path, it will be treated as relative to the default path.
		'''
		if isinstance(d, str):
			isAbsPath = os.path.isabs(d)
			self._slotLoc = d if isAbsPath else os.path.join(self.defaultDir, d) #if the given dir is not a full path, treat it as relative to the default path.
		else: #store object.
			self._slotLoc = d


	@property
	def ui(self) -> QtWidgets.QWidget:
		'''Get the current ui.

		Return:
			(obj) ui
		'''
		return self.getCurrentUi()


	@ui.setter
	def ui(self, ui) -> None:
		'''Register the uiName in history as current and set slot connections.

		Parameters:
			ui (obj): A previously loaded dynamic ui object.
		'''
		self.setCurrentUi(ui)


	@property
	def prevUi(self) -> QtWidgets.QWidget:
		'''Get the previous ui from history.

		Return:
			(obj)
		'''
		return self.getPrevUi()


	@property
	def prevCommand(self) -> object:
		'''Get the last called slot method.

		Return:
			(obj) method.
		'''
		try:
			return self.prevCommands[-1]

		except IndexError as error:
			return None


	@property
	def prevCommands(self) -> tuple:
		'''Get a list of previously called slot methods.

		Return:
			(tuple) list of methods.
		'''
		cmds = [w.getSlot() for w in self._wgtHistory[-10:]] #limit to last 10 elements and get methods from widget history.
		hist = tuple(dict.fromkeys(cmds[::-1]))[::-1] #remove any duplicates (keeping the last element). [hist.remove(l) for l in hist[:] if hist.count(l)>1] #

		return hist


	def initWidgets(self, ui, widgets, recursive=True, returnAllWidgets=False, **kwargs):
		"""Add widgets as attributes of the ui while giving additional attributes to the widgets themselves.

		Parameters:
			ui (obj): A previously loaded dynamic ui object.
			widgets (obj)(list): A widget or list of widgets to be added.
			recursive (bool): Whether to recursively add child widgets (default=True).
			kwargs (): Keyword arguments to set additional widget attributes.

		Return:
			(set) The added widgets.
		"""
		added_widgets = set()
		for w in Iter.makeList(widgets):
			if w in ui._widgets or w in added_widgets or not self.isWidget(w):
				continue

			w.ui = ui
			w.name = w.objectName()
			w.type = w.__class__.__name__
			w.derivedType = self.getDerivedType(w, name=True)
			w.signals = self.getDefaultSignals(w) #default widget signals as list. ie. [<widget.valueChanged>]
			w.prefix = self.getprefix(w.name) #returns an string alphanumberic prefix if name startswith a series of alphanumberic charsinst is followed by three integers. ie. 'cmb' from 'cmb015'
			w.getSlot = lambda w=w, u=ui: getattr(self.getSlots(u), w.name, None)

			if ui.isSubmenu and not w.prefix=='i':
				self.setStyle(w, style=self.submenuStyle, alpha=0)
			else:
				self.setStyle(w, style=self.style, alpha=0.01)

			setAttributes(w, **kwargs)
			setattr(ui, w.name, w)
			added_widgets.add(w)
			# print (0, 'initWidgts:', w.ui.name.ljust(26), w.prefix.ljust(25), (w.name or type(w).__name__).ljust(25), w.type.ljust(15), w.derivedType.ljust(15), id(w)) #debug

			if recursive:
				child_widgets = w.findChildren(QtWidgets.QWidget)
				self.initWidgets(ui, child_widgets, **kwargs)

		ui._widgets.update(added_widgets)
		return added_widgets if not returnAllWidgets else ui._widgets


	@staticmethod
	def getPropertyFromUiFile(file, prop):
		'''Get sub-properties and their values from a given property name.
		Returns all items between the opening and closing statements of the given property.

		So 'customwidget' would return:
			[('class', 'PushButtonDraggable'), ('extends', 'QPushButton'), ('header', 'widgets.pushbuttondraggable.h')]
		from:
			<customwidget>
				<class>PushButtonDraggable</class>
				<extends>QPushButton</extends>
				<header>widgets.pushbuttondraggable.h</header>
			</customwidget>

		Parameters:
			file (str): The full file path to the ui file.
			prop (str): the property text without its opening and closing brackets. 
								ie. 'customwidget' for the property <customwidget>.
		Return:
			(list) list of tuples (typically one or two element).

		Example: getPropertyFromUiFile(file, 'customwidget')
		'''
		f = open(file); # print (f.read()) #debug
		content = list(f.readlines())

		result = []
		actual_prop_text=''
		for i, l in enumerate(content):

			if l.strip() == '<{}>'.format(prop):
				actual_prop_text = l
				start = i+1
			elif l == Str.insert(actual_prop_text, '/', '<'):
				end = i

				delimiters = '</', '<', '>', '\n', ' ',
				regex_pattern = '|'.join(map(re.escape, delimiters))

				lst = [tuple(dict.fromkeys([i for i in re.split(regex_pattern, s) if i])) 
							for s in content[start:end]] #use dict to remove any duplicates.
				result.append(lst)

		f.close()
		return result


	@staticmethod
	def _getUiLevelFromDir(filePath):
		'''Get the UI level by looking for trailing intergers in it's dir name.
		If none are found a default level of 3 (main menu) is used.

		manu types:
			level 0: stackedwidget: start (root)
			level 1: stackedwidget: base menus
			level 2: stackedwidget: sub menus
			level 3: main menus
			level 4: popup menus

		Parameters:
			filePath (str): The directory containing the ui file. ie. 'O:/Cloud/Code/_scripts/uitk/uitk/ui/uiLevel_0/init.ui'

		Return:
			(int) If no level is found, a level of 3 (main menu) will be returned.
		'''
		uiFolder = File.formatPath(filePath, 'dir')
		try:
			return int(re.findall(r"\d+\s*$", uiFolder)[0]) #get trailing integers.
		except IndexError: #not an int.
			return 3


	@staticmethod
	def getprefix(widget):
		'''Return the prefix of a widget's object name.
		A valid prefix is an alphanumeric character sequence at the beginning of the widget's object name,
		ending at the first digit.

		Parameters:
			widget (str)(obj): The widget or its object name as a string.

		Return:
			(str) The prefix of the widget's object name as a string.

		Example:
			getprefix('someName00') #returns: 'someName'
			getprefix('someName') #returns: 'someName'
			getprefix('some0Name') #returns: 'some'
		'''
		import re

		if not isinstance(widget, str):
			widget = widget.objectName()

		match = re.search(r'^\D*', widget)
		return widget[:match.end()]


	@staticmethod
	def _getWidgetsFromUi(ui: QtWidgets.QWidget, inc=[], exc='_*', objectNamesOnly=False) -> dict:
		'''Find widgets in a PySide2 UI object.

		Parameters:
			ui (obj): A previously loaded dynamic ui object.
			inc (str)(tuple): Widget names to include.
			exc (str)(tuple): Widget names to exclude.
			objectNamesOnly (bool): Only include widgets with object names.

		Return:
			(dict) {<widget>:'objectName'}
		'''
		dct = {c:c.objectName() for c in ui.findChildren(QtWidgets.QWidget, None) 
			if (not objectNamesOnly or c.objectName())}

		return Iter.filterDict(dct, inc, exc, keys=True, values=True)


	@staticmethod
	def _getWidgetFromUi(ui: QtWidgets.QWidget, object_name: str) -> QtWidgets.QWidget:
		'''Find a widget in a PySide2 UI object by its object name.

		Parameters:
			ui (obj): A previously loaded dynamic ui object.
			object_name (str): The object name of the widget to find.

		Return:
			(obj)(None) The widget object if it's found, or None if it's not found.
		'''
		return ui.findChild(QtWidgets.QWidget, object_name)


	def _getWidgetsFromDir(self, path):
		'''Get all widget class objects from a given directory, module filepath or module name.

		Parameters:
			path (str): A directory, fullpath to a widget module, or the name of a module residing in the 'widgets' directory.
						For example: - 'path_to/uitk/widgets'
									 - 'path_to/uitk/widgets/comboBox.py'
									 - 'comboBox'
		Return:
			(dict) keys are widget names and the values are the widget class objects.
			Returns an empty dictionary if no widgets were found or an error occurred.
		'''
		mod_name = File.formatPath(path, 'name')
		if mod_name:
			wgt_name = Str.setCase(mod_name, 'pascal')
			if wgt_name in self._registeredWidgets:
				return {}

			if not File.isValidPath(path):
				path = os.path.join(self.widgets, f'{mod_name}.py')

			try:
				spec = importlib.util.spec_from_file_location(mod_name, path)
				mod = importlib.util.module_from_spec(spec)
				spec.loader.exec_module(mod)
			except ModuleNotFoundError as error:
				print(f"# Error: {__file__} in _getWidgetsFromDir\n#\t{error}")
				return {}

			cls_members = inspect.getmembers(mod, lambda m: inspect.isclass(m) 
				and m.__module__==mod.__name__ and issubclass(m, QtWidgets.QWidget)) #get only the widget classes that are defined in the module and not any imported classes.
			return dict(cls_members)

		else: #get all widgets in the given path by recursively calling this fuction.
			try:
				files = os.listdir(path)
			except FileNotFoundError as error:
				print(f"# Error: {__file__} in _getWidgetsFromDir\n#\t{error}")
				return {}

			widgets={}
			for file in files:
				if file.endswith('.py') and not file.startswith('_'):
					mod_path = os.path.join(path, file)
					widgets.update(self._getWidgetsFromDir(mod_path))
			return widgets


	def setSlots(self, ui, clss):
		'''Explicitly set the slot class for a given ui.

		Parameters:
			ui (obj): A previously loaded dynamic ui object.
			clss (obj): A class or class instance to set as the slots location for the given ui. 

		Return:
			(obj) class instance.
		'''
		if inspect.isclass(clss):
			setattr(clss, 'sb', self)
			ui._slots = clss() if callable(clss) else clss
			self.initWidgets(ui, ui.findChildren(QtWidgets.QWidget))
			return ui._slots
		return None


	def getSlots(self, ui, persist=False):
		'''Get the class instance of the ui's slots module if it exists.

		Parameters:
			ui (obj): A previously loaded dynamic ui object.
			persist (bool): Do not assign the ui's slots attribute a None value when a class is not found.

		Return:
			(obj) class instance.
		'''
		if hasAttribute(ui, '_slots'):
			# print ('getSlots:', ui.name, ui._slots, self.slots, inspect.isclass(ui._slots)) #debug
			return ui._slots

		else:
			if isinstance(self.slots, str):
				clss = self._importSlots(ui)
			else: #if slots is a class object:
				clss = self.slots
			# print ('getSlots:2', ui.name, clss, inspect.isclass(clss), self.slots) #debug
			if not clss:
				if ui.isSubmenu:
					mainmenu = self.getUi(ui, 3)
					basemenu = self.getUi(ui, 1)
					if mainmenu: #is a submenu that has a parent menu.
						return self.getSlots(mainmenu)
					elif basemenu:
						return self.getSlots(basemenu)
		if clss:
			return self.setSlots(ui, clss)
		elif not persist:
			ui._slots = None
		return None


	def _importSlots(self, ui, path=None, recursive=True):
		'''Import the slot class associated with the given ui.
		ie. get <Polygons> class using ui name 'polygons'

		Parameters:
			ui (obj): A previously loaded dynamic ui object.
			path (str): The directory location of the module.
					If None is given, the slots property will be used.
			recursive (bool): Search subfolders of the given path.

		Return:
			(obj) class
		'''
		assert isinstance(ui, QtWidgets.QMainWindow), 'Incorrect datatype: {}: {}'.format(type(ui), ui)

		d = File.formatPath(self.slots, 'dir')
		suffix = '' if d=='slots' else d

		mod_name = '{}_{}'.format(
							Str.setCase(ui.name, case='camel'), 
							Str.setCase(suffix, case='camel'),
							).rstrip('_') #ie. 'polygons_maya' or 'polygons' if suffix is None.
		if not path:
			if isinstance(self.slots, str):
				path = self.slots
			else:
				path = self.defaultDir
		path = next((f for f in glob.iglob('{}/**/{}.py'.format(path, mod_name), recursive=recursive)), None)
		if path:
			try: #import the module and return the slots class object.
				spec = importlib.util.spec_from_file_location(mod_name, path) #mod_name: ie. 'polygons_maya'. path: ie. 'full\path\to\maya/polygons_maya.py'
				mod = importlib.util.module_from_spec(spec)
				sys.modules[spec.name] = mod
				spec.loader.exec_module(mod)

				cls_name = Str.setCase(mod_name, case='pascal')
				try:
					clss = getattr(mod, cls_name)
				except AttributeError as error: #if a class by the same name as the module doesn't exist: (ex. <Polygons_maya> from module 'polygons_maya')
					clsmembers = inspect.getmembers(mod, inspect.isclass)
					clss = clsmembers[0][1] #just get the first class.
					print (f'# Warning: {__file__} in _importSlots\n#\t{error}.\n#\tUsing {clss} instead.')
				return clss

			except (FileNotFoundError, ModuleNotFoundError) as error:
				print (f'# Error: {__file__} in _importSlots\n#\t{error}.')
		return None


	def registerWidgets(self, widgets):
		'''Register any custom widgets using the module names.
		Registered widgets can be accessed as properties. ex. sb.PushButton()

		Parameters:
			widgets (str)(obj)(list): A filepath to a dir containing widgets or to the widget itself. 
						ie. 'O:/Cloud/Code/_scripts/uitk/uitk/ui/widgets' or the widget(s) themselves. 

		Return:
			(obj)(list) list if widgets given as a list.

		Example: registerWidgets(<class 'widgets.menu.Menu'>) #register using widget class object.
		Example: registerWidgets('O:/Cloud/Code/_scripts/uitk/uitk/ui/widgets/menu.py') #register using path to widget module.
		'''
		result=[]
		for w in Iter.makeList(widgets): #assure widgets is a list.

			if isinstance(w, str):
				widgets_ = self._getWidgetsFromDir(w)
				for w_ in widgets_.values():
					rw = self.registerWidgets(w_)
					result.append(rw)
				continue

			elif w.__name__ in self._registeredWidgets:
				continue

			try:
				self.registerCustomWidget(w)
				self._registeredWidgets[w.__name__] = w
				setattr(self, w.__name__, w)
				result.append(w)

			except Exception as error:
				print (f'# Error: {__file__} in registerWidgets\n#\t{error}.')

		return Iter.formatReturn(result, widgets) #if 'widgets' is given as a list; return a list.


	def loadAllUi(self, path=None, widgets=None):
		'''Extends the 'loadUi' method to load all ui from a given path.

		Parameters:
			path (str): The path to the directory containing the ui files to load.
				If no path is given all ui from the default 'dynui' will be loaded.
			widgets (str)(obj)(list): A filepath to a dir containing widgets or to the widget itself.
						ie. 'O:/Cloud/Code/_scripts/uitk/uitk/ui/widgets' or the widget(s) themselves.

		Return:
			(list) QMainWindow object(s).
		'''
		files = glob.iglob('{}/**/*.ui'.format(path or self.dynui), recursive=True)
		return [self.loadUi(f, widgets) for f in files]


	def loadUi(self, file, widgets=None):
		'''Loads a ui from the given path to the ui file.

		Parameters:
			file (str): The full file path to the ui file.
			widgets (str)(obj)(list): A filepath to a dir containing widgets or the widget(s) itself.
						ie. 'O:/Cloud/Code/_scripts/uitk/uitk/ui/widgets' or the widget(s) themselves.
		Return:
			(obj) QMainWindow object.
		'''
		name = File.formatPath(file, 'name')
		path = File.formatPath(file, 'path')
		level = self._getUiLevelFromDir(file)

		#register custom widgets
		if widgets is None and not isinstance(self.widgets, str): #widget objects defined in widgets.
			widgets = self.widgets
		if widgets is not None: #widgets given explicitly or defined in widgets.
			self.registerWidgets(widgets)
		else: #search for and attempt to load any widget dependancies using the path defined in widgets.
			lst = self.getPropertyFromUiFile(file, 'customwidget')
			for l in lst: #get any custom widgets from the ui file.
				try:
					className = l[0][1] #ie. 'PushButtonDraggable' from ('class', 'PushButtonDraggable')
					derivedType = l[1][1] #ie. 'QPushButton' from ('extends', 'QPushButton')
				except IndexError as error:
					continue

				mod_name = Str.setCase(className, 'camel')
				fullpath = os.path.join(self.widgets, mod_name+'.py')
				self.registerWidgets(fullpath)

		ui = MainWindow(self, file)
		self._loadedUi.add(ui)
		return ui


	def getUi(self, ui=None, level=None):
		'''Get a dynamic ui using its string name, or if no argument is given, return the current ui.

		Parameters:
			ui (str)(obj)(list): The ui or name(s) of the ui.
			level (int)(list): Integer(s) representing the level to include.
						ex. 2 for submenu, 3 for main menu, or [2, 3] for both.
		Return:
			(str)(list) list if 'level' given as a list.
		'''
		error_msg = lambda ui: f'# Error: {__file__} in getUi\n#\tUI not found: {ui}({type(ui).__name__})\n#\tConfirm the following UI path is correct: {self.dynui}'

		if not ui:
			ui = self.ui

		if level:
			if isinstance(ui, str):
				ui, *tags = ui.split('#')

				ui1 = self.getUi(ui)
				if not ui1 or level==2:
					submenu_name = '{}_{}#{}'.format(ui, 'submenu', '#'.join(tags)).rstrip('#') #reformat as submenu w/tags. ie. 'polygons_submenu#edge' from 'polygons'
					ui1 = self.getUi(submenu_name) #in the case where a submenu exist without a parent menu.
					if not ui1:
						print (error_msg(ui))
						return None

				ui = [u for u in self._loadedUi
						if all([
							u.base==ui1.base, 
							u.tags==Iter.makeList(tags), 
							u.level in Iter.makeList(level),
						])
				]

			else:
				ui = [u for u in self._loadedUi 
						if all([
							u.base==ui.base, 
							u.level in Iter.makeList(level),
						])
				]

			return Iter.formatReturn(ui, level) #if 'level' is given as a list; return a list.

		elif isinstance(ui, str):
			try:
				if not '_submenu' in ui:
					ui, *tags = ui.split('#')

				return getattr(self, ui)

			except AttributeError as error:
				print (error_msg(ui))
				return None

		elif isinstance(ui, (list, set, tuple)):
			return [self.getUi(u) for u in ui]

		else:
			return ui


	def showAndConnect(self, ui, connectOnShow=True):
		'''Register the uiName in history as current, 
		set slot connections, and show the given ui.
		An override for the built-in show method.

		Parameters:
			ui (obj): A previously loaded dynamic ui object.
			connectOnShow (bool): The the ui as connected.
		'''
		if connectOnShow:
			ui.connected = True
		ui.__class__.show(ui)


	def getCurrentUi(self) -> object:
		'''Get the current ui.

		Return:
			(obj) ui
		'''
		try:
			return self._currentUi

		except AttributeError as error:
			if len(self._loadedUi)==1: #if only one ui is loaded set that ui as current.
				ui = self._loadedUi.pop()
				self.setCurrentUi(ui)
				return ui
			elif self.dynui.endswith('.ui'): #if the ui location is set to a single ui, then load and set that ui as current.
				ui = self.loadUi(self.dynui)
				self.setCurrentUi(ui)
				return ui

			return None


	def setCurrentUi(self, ui) -> None:
		'''Register the specified dynamic UI as the current one in the application's history.
		Once registered, the `ui` object can be accessed through the `ui` property while it remains as the current UI.
		If the specified `ui` is already the current UI, the method simply returns without making any changes. 

		Parameters:
			ui (obj): A previously loaded dynamic ui object.
		'''
		try:
			if self._currentUi==ui:
				return
		except AttributeError as error:
			pass

		# print ('set current ui:', ui.name, id(self), locals()) #debug
		self._currentUi = ui
		self._uiHistory.append(ui)
		# print ('_uiHistory:', [u.name for u in self._uiHistory]) #debug


	def getPrevUi(self, allowDuplicates=False, allowCurrent=False, omitLevel=None, asList=False):
		'''Get ui from history.
		ex. _uiHistory list: ['previousName2', 'previousName1', 'currentName']

		Parameters:
			allowDuplicates (bool): Applicable when returning asList. Allows for duplicate names in the returned list.
			omitLevel (int)(list): Remove instances of the given ui level(s) from the results. Default is [] which omits nothing.
			allowCurrent (bool): Allow the currentName. Default is off.
			asList (bool): Returns the full list of previously called names. By default duplicates are removed.

		Return:
			(str)(list) if 'asList': returns [list of string names]
		'''
		self._uiHistory = self._uiHistory[-200:] #keep original list length restricted to last 200 elements
		hist = self._uiHistory.copy() #work on a copy of the list, keeping the original intact

		if not allowCurrent:
			hist = hist[:-1] #remove the last index. (currentName)

		if not allowDuplicates:
			[hist.remove(u) for u in hist[:] if hist.count(u)>1] #remove any previous duplicates if they exist; keeping the last added element.

		if omitLevel is not None:
			hist = [u for u in hist if not u.level in Iter.makeList(omitLevel)] #remove any items having a ui level of those in the omitLevel list.

		if asList:
			return hist #return entire list after being modified by any flags such as 'allowDuplicates'.
		else:
			try:
				return hist[-1] #return the previous ui name if one exists.
			except:
				return None


	@staticmethod
	def getDerivedType(widget, name=False, module='QtWidgets', inc=[], exc=[], filterByBaseType=False):
		'''Get the base class of a custom widget.
		If the type is a standard widget, the derived type will be that widget's type.

		Parameters:
			widget (str)(obj): QWidget or it's objectName.
			name (bool): Return the class or the class name.
			module (str): The name of the base class module to check for.
			inc (list): Widget types to include. All other will be omitted. Exclude takes dominance over include. Meaning, if the same attribute is in both lists, it will be excluded.
			exc (list): Widget types to exclude. ie. ['QWidget', 'QAction', 'QLabel', 'QPushButton', 'QListWidget']
			filterByBaseType (bool): When using `inc`, or `exc`; Filter by base class name, or derived class name. ie. 'QLayout'(base) or 'QGridLayout'(derived)

		Return:
			(obj)(string)(None) class or class name if `name`. ie. 'QPushButton' from a custom widget with class name: 'PushButton'
		'''
		# print(widget.__class__.__mro__) #debug
		for c in widget.__class__.__mro__:
			if c.__module__==module or c.__module__.split('.')[-1]==module: #check for the first built-in class. ie. 'PySide2.QtWidgets' or 'QtWidgets'
				typ = c.__class__.__base__.__name__ if filterByBaseType else c
				if not (typ in exc and (typ in inc if inc else typ not in inc)):
					return typ.__name__ if name else typ


	def getWidget(self, name, ui=None):
		'''Case insensitive. Get the widget object/s from the given ui and name.

		Parameters:
			name (str): The object name of the widget. ie. 'b000'
			ui (str)(obj): ui, or name of ui. ie. 'polygons'. If no nothing is given, the current ui will be used.
							A ui object can be passed into this parameter, which will be used to get it's corresponding name.
		Return:
			(obj) if name:  widget object with the given name from the current ui.
				  if ui and name: widget object with the given name from the given ui name.
			(list) if ui: all widgets for the given ui.
		'''
		if ui is None or isinstance(ui, str):
			ui = self.getUi(ui)

		return next((w for w in ui.widgets if w.name==name), None)


	def getWidgetsByType(self, types, ui=None, derivedType=False):
		'''Get widgets of the given types.

		Parameters:
			types (str)(list): A widget class name, or list of widget class names. ie. 'QPushbutton' or ['QPushbutton', 'QComboBox']
			ui (str)(obj): Parent ui name, or ui object. ie. 'polygons' or <polygons>
							If no name is given, the current ui will be used.
			derivedType (bool): Get by using the parent class of custom widgets.

		Return:
			(list)
		'''
		if ui is None or isinstance(ui, str):
			ui = self.getUi(ui)

		typ = 'derivedType' if derivedType else 'type'
		return [w for w in ui.widgets if getattr(w, typ) in Iter.makeList(types)]


	def getWidgetName(self, widget=None, ui=None):
		'''Get the widget's stored string objectName.

		Parameters:
			widget (obj): The widget to get the object name of.
					If no widget is given, names of all widgets will be returned.
			ui (str)(obj): The parent ui, or ui name. ie. <polygons> or 'polygons'
					If no name is given, the current ui will be used.
		Return:
			(str)(list)
			if widget: (str) the widget objectName for the given widget.
			if ui: the widget objectNames for widgets of the given ui name.
			if not ui: the widget objectNames for widgets of the current ui.
		'''
		if not isinstance(ui, QtWidgets.QMainWindow):
			ui = self.getUi(ui)

		if widget:
			return widget.name

		else: #return all objectNames from ui name.
			return [w.name for w in ui.widgets]


	def getWidgetFromMethod(self, method):
		'''Get the corresponding widget from a given method.

		Parameters:
			method (obj): The method in which to get the widget of.

		Return:
			(obj) widget. ie. <b000 widget> from <b000 method>.
		'''
		if not method:
			return None

		return next(iter(w for u in self._loadedUi for w in u.widgets if w.getSlot()==method), None)


	def getMethod(self, ui, widget=None):
		'''Get the method(s) associated with the given ui / widget.

		Parameters:
			ui (str)(obj): The ui name, or ui object. ie. 'polygons' or <polygons>
			widget (str)(obj): widget, widget's objectName, or method name.

		Return:
			if widget: corresponding method object to given widget.
			else: all of the methods associated to the given ui name as a list.

		Example:
			sb.getSlot('polygons', <b022>)() #call method <b022> of the 'polygons' class
		'''
		if not isinstance(ui, QtWidgets.QMainWindow):
			ui = self.getUi(ui)

		if widget is None: #get all methods for the given ui name.
			return [w.getSlot() for w in ui.widgets]

		elif isinstance(widget, str):
			return next(iter(w.getSlot() for w in ui.widgets if w.getSlot().__name__==widget), None)

		elif not widget in ui._widgets:
			self.initWidgets(ui, widget)

		return next(iter(w.getSlot() for w in ui.widgets if w.getSlot()==widget.getSlot()), None)


	def getSignals(self, w, d=True, exc=[]):
		'''Get all signals for a given widget.

		Parameters:
			w (str)(obj): The widget to get signals for.
			d (bool): Return signals from all derived classes instead of just the given widget class.
				ex. get: QObject, QWidget, QAbstractButton, QPushButton signals from 'QPushButton'
			exc (list): Exclude any classes in this list. ex. exc=[QtCore.QObject, 'QWidget']

		Return:
			(list)

		Example: getSignals(QtWidgets.QPushButton)
			would return:
				clicked (QAbstractButton)
				pressed (QAbstractButton)
				released (QAbstractButton)
				toggled (QAbstractButton)
				customContextMenuRequested (QWidget)
				windowIconChanged (QWidget)
				windowIconTextChanged (QWidget)
				windowTitleChanged (QWidget)
				destroyed (QObject)
				objectNameChanged (QObject)
		'''
		signals=[]
		clss = source if isinstance(source, type) else type(source)
		signal = type(QtCore.Signal())
		for subcls in clss.mro():
			clsname = f'{subcls.__module__}.{subcls.__name__}'
			for k, v in sorted(vars(subcls).items()):
				if isinstance(v, signal):
					if (not d and clsname!=clss.__name__) or (exc and (clss in exc or clss.__name__ in exc)): #if signal is from parent class QAbstractButton and given widget is QPushButton:
						continue
					signals.append(k)
		return signals


	def getDefaultSignals(self, w):
		'''Get the default signals for a given widget type.

		Parameters:
			widgetType (str): Widget class name. ie. 'QPushButton'

		Return:
			(str) signal ie. 'released'
		'''
		signals=[]
		try: #if the widget type has a default signal assigned in the signals dict; get the signal.
			signalTypes = self.defaultSignals[w.derivedType]
			for s in Iter.makeList(signalTypes): #assure 'signalTypes' is a list.
				signal = getattr(w, s, None)
				signals.append(signal)

		except KeyError:
			pass
		return signals


	def setConnections(self, ui):
		'''Replace any signal connections of a previous ui with the set for the ui of the given name.

		Parameters:
			ui (obj): A previously loaded dynamic ui object.
		'''
		assert isinstance(ui, QtWidgets.QMainWindow), 'Incorrect datatype: {}: {}'.format(type(ui), ui)

		prevUi = self.getPrevUi(allowDuplicates=True)
		if prevUi==ui:
			return

		if prevUi and prevUi.isConnected and prevUi.level<3:
			self.disconnectSlots(prevUi)

		if ui.level<3 or not ui.isConnected:
			self.connectSlots(ui)

		submenus = self.getUi(ui, [2]) #give level arg as a list to assure that a list is returned.
		for submenu in submenus:
			mainmenu = self.getUi(ui, 3)
			self.syncAllWidgets(submenu, mainmenu) #sync any submenu widgets with their main menu counterparts.


	def connectSlots(self, ui, widgets=None):
		'''Connect signals to slots for the widgets of the given ui.
		Works with both single slots or multiple slots given as a list.

		Parameters:
			ui (obj): A previously loaded dynamic ui object.
			widgets (obj)(list): QWidget(s)
		'''
		if widgets is None:
			if ui.isConnected:
				return
			widgets = ui.widgets
		# print ('connectSlots:', ui.name, [w.objectName() for w in Iter.makeList(widgets)]) #debug

		for w in Iter.makeList(widgets): #convert 'widgets' to a list if it is not one already.
			# print ('           >:', w.name, w.getSlot()) #debug
			if w.getSlot() and w.signals:
				for s in w.signals:
					if not s:
						continue
					try:
						if isinstance(w.getSlot(), (list, set, tuple)):
							map(s.connect, w.getSlot()) #connect to multiple slots from a list.
						else:
							s.connect(w.getSlot()) #connect single slot (ex. main and cameras ui)

						s.connect(lambda *args, w=w: self._wgtHistory.append(w)) #add the widget to the widget history set on connect. (*args prevents 'w' from being overwritten by the parameter emitted by the signal.)

					except Exception as error:
						print(f'# Error: {__file__} in connectSlots\n#\t{ui.name} {w.name} {s} {w.getSlot()}\n#\t{error}.')

		ui.isConnected = True #set ui state as slots connected.


	def disconnectSlots(self, ui, widgets=None):
		'''Disconnect signals from slots for the widgets of the given ui.
		Works with both single slots or multiple slots given as a list.

		Parameters:
			ui (obj): A previously loaded dynamic ui object.
			widgets (obj)(list): QWidget
		'''
		# print ('disconnectSlots:', ui.name) #debug
		if widgets is None:
			if not ui.isConnected:
				return
			widgets = ui.widgets

		for w in Iter.makeList(widgets):  #convert 'widgets' to a list if it is not one already.
			# print ('           >:', w.name, w.getSlot()) #debug
			if w.getSlot() and w.signals:
				for s in w.signals:
					if not s:
						continue
					try:
						if isinstance(w.getSlot(), (list, set, tuple)):
							s.disconnect() #disconnect all #map(signal.disconnect, slot) #disconnect multiple slots from a list.
						else:
							s.disconnect(w.getSlot()) #disconnect single slot (main and cameras ui)

					except Exception as error:
						print(f'# Error: {__file__} in disconnectSlots\n#\t{ui.name} {w.name} {s} {w.getSlot()}\n#\t{error}.')

		ui.isConnected = False #set ui state as slots disconnected.


	def connect(self, widgets, signals, slots, clss=None):
		'''Connect multiple signals to multiple slots at once.

		Parameters:
			widgets (str)(obj)(list): ie. 'chk000-2' or [tb.ctxMenu.chk000, tb.ctxMenu.chk001]
			signals (str)(list): ie. 'toggled' or ['toggled']
			slots (obj)(list): ie. self.cmb002 or [self.cmb002]
			clss (obj)(list): if the widgets arg is given as a string, then the class it belongs to can be explicitly given. else, the current ui will be used.

		ex call: connect_('chk000-2', 'toggled', self.cmb002, tb.ctxMenu)
		*or connect_([tb.ctxMenu.chk000, tb.ctxMenu.chk001], 'toggled', self.cmb002)
		*or connect_(tb.ctxMenu.chk015, 'toggled', 
				[lambda state: self.rigging.tb004.setText('Unlock Transforms' if state else 'Lock Transforms'), 
				lambda state: self.rigging_submenu.tb004.setText('Unlock Transforms' if state else 'Lock Transforms')])
		'''
		if isinstance(widgets, (str)):
			try:
				widgets = self.getWidgetsFromStr(clss, widgets, showError=True) #getWidgetsFromStr returns a widget list from a string of objectNames.
			except Exception as error:
				widgets = self.getWidgetsFromStr(self.getCurrentUi(), widgets, showError=True)

		#if the variables are not of a list type; convert them.
		widgets = Iter.makeList(widgets)
		signals = Iter.makeList(signals)
		slots = Iter.makeList(slots)

		for widget in widgets:
			for signal in signals:
				signal = getattr(widget, signal)
				for slot in slots:
					signal.connect(slot)


	def syncAllWidgets(self, frm, to, **kwargs):
		'''Extends setSynConnections method to set sync connections 
		for all widgets of the given pair of ui objects.

		Parameters:
			frm (obj): A previously loaded dynamic ui object to sync widgets from.
			to (obj): A previously loaded dynamic ui object to sync widgets to.
		'''
		for w1 in frm.widgets:
			try:
				w2 = self.getWidget(w1.name, to)
			except AttributeError as error:
				continue

			pair_id = hash((w1, w2))
			if pair_id in self._synced_pairs:
				continue

			self.syncWidgets(w1, w2, **kwargs)


	def syncWidgets(self, w1, w2, **kwargs):
		'''Set the initial signal connections that will call the _syncAttributes function on state changes.

		Parameters:
			w1 (obj): The first widget to sync.
			w2 (obj): The second widget to sync.
			kwargs = The attribute(s) to sync as keyword arguments.
		'''
		try:
			signals1 = self.getDefaultSignals(w1) #get the default signal for the given widget.
			signals2 = self.getDefaultSignals(w2)

			for s1, s2 in zip(signals1, signals2): #set sync connections for each of the widgets signals.
				s1.connect(lambda: self._syncAttributes(w1, w2, **kwargs))
				s2.connect(lambda: self._syncAttributes(w2, w1, **kwargs))

			pair_id = hash((w1, w2))
			self._synced_pairs.add(pair_id)

		except (AttributeError, KeyError) as error:
			# if w1 and w2: print ('# {}: {}.syncWidgets({}, {}): {}. args: {}, {} #'.format('KeyError' if type(error)==KeyError else 'AttributeError', __name__, w1.objectName(), w2.objectName(), error, w1, w2)) #debug
			return


	attributesGetSet = {
		'value':'setValue', 
		'text':'setText', 
		'icon':'setIcon', 
		'checkState':'setCheckState', 
		'isChecked':'setChecked', 
		'isDisabled':'setDisabled', 
	}
	def _syncAttributes(self, frm, to, attributes=[]):
		'''Sync the given attributes between the two given widgets.
		If a widget does not have an attribute it will be silently skipped.

		Parameters:
			frm (obj): The widget to transfer attribute values from.
			to (obj): The widget to transfer attribute values to.
			attributes (str)(list)(dict): The attribute(s) to sync. ie. a setter attribute 'setChecked' or a dict containing getter:setter pairs. ie. {'isChecked':'setChecked'}
		'''
		if not attributes:
			attributes = self.attributesGetSet

		elif not isinstance(attributes, dict):
			attributes = {next((k for k,v in self.attributesGetSet.items() if v==i), None):i #construct a gettr setter pair dict using only the given setter values.
				 for i in Iter.makeList(attributes)
			}

		_attributes = {}
		for gettr, settr in attributes.items():
			try:
				_attributes[settr] = getattr(frm, gettr)()
			except AttributeError as error:
				pass

		for attr, value in _attributes.items(): #set the second widget's attributes from the first.
			try:
				getattr(to, attr)(value)
			except AttributeError as error:
				pass


	def setWidgetAttrs(self, *args, **kwargs):
		'''Set multiple properties, for multiple widgets, on multiple ui's at once.

		Parameters:
			*args = arg [0] (str) String of objectNames. - objectNames separated by ',' ie. 'b000-12,b022'
					arg [1:] dynamic ui object/s.  If no ui's are given, then the parent and child uis will be used.
			*kwargs = keyword: - the property to modify. ex. setText, setValue, setEnabled, setDisabled, setVisible, setHidden
					value: - intended value.

		Example:
			setWidgetAttrs('chk003', <ui1>, <ui2>, setText='Un-Crease')
		'''
		if not args[1:]:
			mainmenu = self.getUi(self.getCurrentUi(), 3)
			submenu = self.getUi(self.getCurrentUi(), 2)
			args = args+(mainmenu, submenu)

		for ui in args[1:]:
			widgets = self.getWidgetsFromStr(ui, args[0]) #getWidgetsFromStr returns a widget list from a string of objectNames.
			for property_, value in kwargs.items():
				[getattr(w, property_)(value) for w in widgets] #set the property state for each widget in the list.


	@staticmethod
	def unpackNames(nameString):
		'''Get a list of individual names from a single name string.
		If you are looking to get multiple objects from a name string, call 'getWidgetsFromStr' directly instead.

		Parameters:
			nameString = string consisting of widget names separated by commas. ie. 'v000, b004-6'

		Return:
			unpacked names. ie. ['v000','b004','b005','b006']

		Example: unpackNames('chk021-23, 25, tb001')
		'''
		packed_names = [n.strip() for n in nameString.split(',') #build list of all widgets passed in containing '-' 
							if '-' in n or n.strip().isdigit()]

		otherNames = [n.strip() for n in nameString.split(',') #all widgets passed in not containing '-'
							if '-' not in n and not n.strip().isdigit()]

		unpacked_names=[] #unpack the packed names:
		for name in packed_names:
			if '-' in name:
				name = name.split('-') #ex. split 'b000-8'
				prefix = name[0].strip('0123456789') #ex. split 'b' from 'b000'
				start = int(name[0].strip('abcdefghijklmnopqrstuvwxyz') or 0) #start range. #ex. '000' #converting int('000') returns None, if case; assign 0.
				stop = int(name[1])+1 #end range. #ex. '8' from 'b000-8' becomes 9, for range up to 9 but not including 9.
				unpacked_names.extend([str(prefix)+'000'[:-len(str(num))]+str(num) for num in range(start,stop)]) #build list of name strings within given range
				last_name = name
				last_prefix = prefix
			else:
				num = name
				unpacked_names.extend([str(last_prefix)+'000'[:-len(str(num))]+str(num)])

		return otherNames+unpacked_names


	@classmethod
	def getWidgetsFromStr(cls, ui, nameString, showError=False):
		'''Get a list of corresponding widgets from a single shorthand formatted string.
		ie. 's000,b002,cmb011-15' would return object list: [<s000>, <b002>, <cmb011>, <cmb012>, <cmb013>, <cmb014>, <cmb015>]

		Parameters:
			ui (obj): A previously loaded dynamic ui object.
			nameString (str): Widget object names separated by ','. ie. 's000,b004-7'. b004-7 specifies buttons b004-b007.
			showError (bool): Print an error message to the console if a widget is not found.

		Return:
			(list)

		#ex call: getWidgetsFromStr(<ui>, 's000,b002,cmb011-15')
		'''
		widgets=[]
		for n in cls.unpackNames(nameString):
			try:
				w = getattr(ui, n)
				widgets.append(w)
			except AttributeError as error:
				if showError:
					print (f'# Error: {__file__} in getWidgetsFromStr\n#\t{error}.')
				pass

		return widgets


	@staticmethod
	def getCenter(w):
		'''Get the center point of a given widget.

		Parameters:
			w (obj): The widget to query.

		Return:
			(obj) QPoint
		'''
		return QtGui.QCursor.pos() - w.rect().center()


	@staticmethod
	def resizeAndCenterWidget(widget, paddingX=30, paddingY=6):
		'''Adjust the given widget's size to fit contents and re-center.

		Parameters:
			widget (obj): The widget to resize.
			paddingX (int): Any additional width to be applied.
			paddingY (int): Any additional height to be applied.
		'''
		p1 = widget.rect().center()
		widget.resize(widget.sizeHint().width()+paddingX, widget.sizeHint().height()+paddingY)
		p2 = widget.rect().center()
		diff = p1-p2
		widget.move(widget.pos()+diff)


	@staticmethod
	def moveAndCenterWidget(w, p, offsetX=2, offsetY=2):
		'''Move and center the given widget on the given point.

		Parameters:
			w (obj): The widget to resize.
			p (obj): A point to move to.
			offsetX (int): The desired offset on the x axis. 2 is center. 
			offsetY (int): The desired offset on the y axis.
		'''
		width = p.x()-(w.width()/offsetX)
		height = p.y()-(w.height()/offsetY)

		w.move(QtCore.QPoint(width, height)) #center a given widget at a given position.


	@staticmethod
	def centerWidgetOnScreen(w):
		'''
		'''
		centerPoint = QtGui.QScreen.availableGeometry(QtWidgets.QApplication.primaryScreen()).center()
		w.move(centerPoint - w.frameGeometry().center())


	def toggleWidgets(self, *args, **kwargs):
		'''Set multiple boolean properties, for multiple widgets, on multiple ui's at once.

		Parameters:
			*args = dynamic ui object/s. If no ui's are given, then the current UI will be used.
			*kwargs = keyword: - the property to modify. ex. setChecked, setUnChecked, setEnabled, setDisabled, setVisible, setHidden
						value: string of objectNames - objectNames separated by ',' ie. 'b000-12,b022'
		Example:
			toggleWidgets(<ui1>, <ui2>, setDisabled='b000', setUnChecked='chk009-12', setVisible='b015,b017')
		'''
		if not args:
			mainmenu = self.getUi(self.getCurrentUi(), 3)
			submenu = self.getUi(self.getCurrentUi(), 2)
			args = [submenu, mainmenu]

		for ui in args:
			for k in kwargs: #property_ ie. setUnChecked
				widgets = self.getWidgetsFromStr(ui, kwargs[k]) #getWidgetsFromStr returns a widget list from a string of objectNames.

				state = True
				if 'Un' in k: #strips 'Un' and sets the state from True to False. ie. 'setUnChecked' becomes 'setChecked' (False)
					k = k.replace('Un', '')
					state = False

				[getattr(w, k)(state) for w in widgets] #set the property state for each widget in the list.


	def setAxisForCheckBoxes(self, checkboxes, axis, ui=None):
		'''Set the given checkbox's check states to reflect the specified axis.

		Parameters:
			checkboxes (str)(list): 3 or 4 (or six with explicit negative values) checkboxes.
			axis (str): Axis to set. Valid text: '-','X','Y','Z','-X','-Y','-Z' ('-' indicates a negative axis in a four checkbox setup)

		ex call: setAxisForCheckBoxes('chk000-3', '-X') #optional ui arg for the checkboxes
		'''
		if isinstance(checkboxes, (str)):
			if ui is None:
				ui = self.getCurrentUi()
			checkboxes = self.getWidgetsFromStr(ui, checkboxes)

		prefix = '-' if '-' in axis else '' #separate the prefix and axis
		coord = axis.strip('-')

		for chk in checkboxes:
			if any([chk.text()==prefix, chk.text()==coord, chk.text()==prefix+coord]):
				chk.setChecked(True)


	def getAxisFromCheckBoxes(self, checkboxes, ui=None):
		'''Get the intended axis value as a string by reading the multiple checkbox's check states.

		Parameters:
			checkboxes (str)(list): 3 or 4 (or six with explicit negative values) checkboxes. Valid text: '-','X','Y','Z','-X','-Y','-Z' ('-' indicates a negative axis in a four checkbox setup)

		Return:
			(str) axis value. ie. '-X'		

		ex call: getAxisFromCheckBoxes('chk000-3')
		'''
		if isinstance(checkboxes, (str)):
			if ui is None:
				ui = self.getCurrentUi()
			checkboxes = self.getWidgetsFromStr(ui, checkboxes, showError=True)

		prefix=axis=''
		for chk in checkboxes:
			if chk.isChecked():
				if chk.text()=='-':
					prefix = '-'
				else:
					axis = chk.text()
		# print ('prefix:', prefix, 'axis:', axis) #debug
		return prefix+axis #ie. '-X'


	def gcProtect(self, obj=None, clear=False):
		'''Protect the given object from garbage collection.

		Parameters:
			obj (obj)(list): The obj(s) to add to the protected list.
			clear (bool): Clear the set before adding any given object(s).

		Return:
			(list) protected objects.
		'''
		if clear:
			self._gcProtect.clear()

		for o in Iter.makeList(obj):
			self._gcProtect.add(o)

		return self._gcProtect


	def isWidget(self, obj):
		'''Returns True if the given obj is a valid widget.

		Parameters:
			obj (obj): An object to query.

		Return:
			(bool)
		'''
		try:
			return issubclass(obj, QtWidgets.QWidget)
		except TypeError:
			return issubclass(obj.__class__, QtWidgets.QWidget)


	@staticmethod
	def getWidgetAt(pos, topWidgetOnly=True):
		'''Get visible and enabled widget(s) located at the given position.
		As written, this will disable `TransparentForMouseEvents` on each widget queried.

		Parameters:
			pos (QPoint) = The global position at which to query.
			topWidgetOnly (bool): Return only the top-most widget, 
				otherwise widgets are returned in the order in which they overlap.
				Disabling this option will cause overlapping windows to flash as 
				their attribute is changed and restored.
		Return:
			(obj)(list) list if not topWidgetOnly.

		Example: getWidgetAt(QtGui.QCursor.pos())
		'''
		w = QtWidgets.QApplication.widgetAt(pos)
		if topWidgetOnly:
			return w

		widgets=[]
		while w:
			widgets.append(w)

			w.setAttribute(QtCore.Qt.WA_TransparentForMouseEvents) #make widget invisible to further enquiries.
			w = QtWidgets.QApplication.widgetAt(pos)

		for w in widgets: #restore attribute.
			w.setAttribute(QtCore.Qt.WA_TransparentForMouseEvents, False)

		return widgets


	@staticmethod
	def getParentWidgets(widget, objectNames=False):
		'''Get the all parent widgets of the given widget.

		Parameters:
			widget (obj): QWidget
			objectNames (bool): Return as objectNames.

		Return:
			(list) Object(s) or objectName(s)
		'''
		parentWidgets=[]
		w = widget
		while w:
			parentWidgets.append(w)
			w = w.parentWidget()
		if objectNames:
			return [w.objectName() for w in parentWidgets]
		return parentWidgets


	@staticmethod
	def getTopLevelParent(widget, index=-1):
		'''Get the parent widget at the top of the hierarchy for the given widget.

		Parameters:
			widget (obj): QWidget
			index (int): Last index is top level.

		Return:
			(QWidget)
		'''
		return self.getParentWidgets()[index]


	@staticmethod
	def getAllWindows(name=None):
		'''Get Qt windows.

		Parameters:
			name (str): Return only windows having the given object name.

		Return:
			(list) windows.
		'''
		return [w for w in QtWidgets.QApplication.allWindows() if (name is None) or (w.objectName()==name)]


	@staticmethod
	def getAllWidgets(name=None):
		'''Get Qt widgets.

		Parameters:
			name (str): Return only widgets having the given object name.

		Return:
			(list) widgets.
		'''
		return [w for w in QtWidgets.QApplication.allWidgets() if (name is None) or (w.objectName()==name)]


	def messageBox(self, string, messageType='', location='topMiddle', timeout=1):
		'''Spawns a message box with the given text.
		Supports HTML formatting.
		Prints a formatted version of the given string to console, stripped of html tags, to the console.

		Parameters:
			messageType (str): The message context type. ex. 'Error', 'Warning', 'Note', 'Result'
			location (str)(point) = move the messagebox to the specified location. Can be given as a qpoint or string value. default is: 'topMiddle'
			timeout (int): time in seconds before the messagebox auto closes.
		'''
		if messageType:
			string = f'{messageType.capitalize()}: {string}'

		try:
			self._messageBox.location = location
		except AttributeError:
			self._messageBox = self.MessageBox(self.parent())
			self._messageBox.location = location
		self._messageBox.timeout = timeout

		from re import sub
		print(f"# {sub('<.*?>', '', string)}") #strip everything between '<' and '>' (html tags)

		self._messageBox.setText(string)
		self._messageBox.exec_()


	# @classmethod
	# def progress(cls, fn):
	# 	'''A decorator for progressBar.
	#	Does not work with staticmethods.
	# 	'''
	# 	def wrapper(self, *args, **kwargs):
	# 		self.progressBar(fn(self, *args, **kwargs))
	# 	return wrapper

	def progressBar(self):
		'''
		'''
		try:
			return self._progressBar

		except AttributeError as error:
			from widgets.progressBar import ProgressBar
			self._progressBar = ProgressBar(self.parent())

			try:
				self.ui.progressBar.step1
			except AttributeError:
				pass

			return self._progressBar


	@staticmethod
	def invertOnModifier(value):
		'''Invert a numerical or boolean value if the alt key is pressed.

		Parameters:
			value (int, float, bool) = The value to invert.
		
		Return:
			(int, float, bool)
		'''
		modifiers = QtWidgets.QApplication.instance().keyboardModifiers()
		if not modifiers in (QtCore.Qt.AltModifier, QtCore.Qt.ControlModifier | QtCore.Qt.AltModifier):
			return value

		if type(value) in (int, float):
			result = abs(value) if value<0 else -value
		elif type(value)==bool:
			result = True if value else False

		return result

# --------------------------------------------------------------------------------------------









# --------------------------------------------------------------------------------------------

if __name__=='__main__':

	app = QtWidgets.QApplication.instance() or QtWidgets.QApplication(sys.argv) #return the existing QApplication object, or create a new one if none exists.

	from uitk.slots.polygons import Polygons
	sb = Switchboard(slots='slots/maya') #set relative paths, and explicity set the slots class instead of providing a path like: slots='slots/maya', which in this case would produce the same result with just a little more overhead.
	ui = sb.polygons #get the ui by it's name.
	ui.show()

	print ('current ui:', sb.ui)
	print ('current ui name:', sb.ui.name)
	print ('connected state:', ui.isConnected)
	print ('slots:', ui.slots)
	print ('method tb000:', ui.tb000.getSlot())
	print ('widget from method tb000:', sb.getWidgetFromMethod(ui.tb000.getSlot()))
	# print ('widgets:', ui.widgets)

	exit_code = app.exec_()
	if exit_code != -1:
		sys.exit(exit_code) # run app, show window, wait for input, then terminate program with a status code returned from app.

print (__name__) #module name
# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------


# --------------------------------------------------------------------------------------------
# deprecated:
# --------------------------------------------------------------------------------------------

	# def __getattr__(self, attr_name):
	# 	found_widget = self.sb._getWidgetFromUi(self, attr_name)
	# 	if found_widget:
	# 		self.sb.initWidgets(self, found_widget)
	# 		return found_widget
	# 	raise AttributeError(f'{self.__class__.__name__} has no attribute `{attr_name}`')

# def getprefix(widget):
# 		'''Query a widgets prefix.
# 		A valid prefix is returned when the given widget's objectName startswith an alphanumeric char, 
# 		followed by at least three integers. ex. i000 (alphanum,int,int,int)

# 		Parameters:
# 			widget (str)(obj): A widget or it's object name.

# 		Return:
# 			(str)
# 		'''
# 		prefix=''
# 		if not isinstance(widget, (str)):
# 			widget = widget.objectName()
# 		for char in widget:
# 			if not char.isdigit():
# 				prefix = prefix+char
# 			else:
# 				break

# 		i = len(prefix)
# 		integers = [c for c in widget[i:i+3] if c.isdigit()]
# 		if len(integers)>2 or len(widget)==i:
# 			return prefix

	# def setAttributes(self, obj=None, order=['setVisible'], **kwargs):
	# 	'''Set attributes for a given object.

	# 	Parameters:
	# 		obj (obj): the child obj, or widgetAction to set attributes for. (default=self)
	# 		order (list): List of string keywords. ie. ['move', 'setVisible']. attributes in this list will be set last, in order of the list. an example would be setting move positions after setting resize arguments.
	# 		**kwargs = The keyword arguments to set.
	# 	'''
	# 	if not kwargs:
	# 		return

	# 	obj = obj if obj else self

	# 	for k in order:
	# 		v = kwargs.pop(k, None)
	# 		if v:
	# 			from collections import OrderedDict
	# 			kwargs = OrderedDict(kwargs)
	# 			kwargs[k] = v

	# 	for attr, value in kwargs.items():
	# 		try:
	# 			getattr(obj, attr)(value)

	# 		except AttributeError as error:
	# 			pass; # print (__name__+':','setAttributes:', obj, order, kwargs, error)


# !/usr/bin/python
# coding=utf-8
from uitk.slots.max import *
from uitk.slots.rigging import Rigging



class Rigging_max(Rigging, Slots_max):
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		cmb = self.sb.rigging.draggable_header.ctxMenu.cmb000
		items = ['Bone Tools','Parameter Editor','Parameter Collector','Parameter Wire Dialog']
		cmb.addItems_(items, 'Rigging Editors')

		cmb = self.sb.rigging.cmb001
		items = ['Bones IK Chain','Point','Dummy','Grid','Expose Transform','Lattice','Biped']
		cmb.addItems_(items, "Create")


	def cmb000(self, index=-1):
		'''Editors
		'''
		cmb = self.sb.rigging.draggable_header.ctxMenu.cmb000

		if index>0:
			text = cmb.items[index]
			if text=='Bone Tools':
				maxEval('macros.run "Animation Tools" "BoneAdjustmentTools"')
			elif text=='Parameter Editor':
				maxEval('macros.run "Customize User Interface" "Custom_Attributes"')
			elif text=='Parameter Collector':
				maxEval('macros.run "Parameter Collector" "ParamCollectorShow"')
			elif text=='Parameter Wire Dialog':
				maxEval('macros.run "Parameter Wire" "paramWire_dialog"')
			cmb.setCurrentIndex(0)


	def cmb001(self, index=-1):
		'''Create
		'''
		cmb = self.sb.rigging.cmb001

		if index>0:
			text = cmb.items[index]
			if text=='Bones IK Chain':
				maxEval('macros.run "Inverse Kinematics" "Bones"') #create joint tool
			elif text=='Point':
				maxEval('macros.run "Objects Helpers" "Point"') #Point pos:[46.5545,-11.1098,0] isSelected:on #locator
			elif text=='Dummy':
				maxEval('macros.run "Objects Helpers" "Dummy"')
			elif text=='Grid':
				maxEval('macros.run "Objects Helpers" "Grid"') #grid pos:[14.957,-79.0478,0] isSelected:on width:49.0621 length:51.0787
			elif text=='Expose Transform':
				maxEval('macros.run "Objects Helpers" "ExposeTM"') #ExposeTm pos:[1.12888,-35.9943,0] isSelected:on
			elif text=='Lattice':
				maxEval('modPanel.addModToSelection (Lattice ()) ui:on') #create lattice
			elif text=='macros.run "Objects Systems" "Biped"':
				maxEval('macros.run "Objects Systems" "Biped"')
			cmb.setCurrentIndex(0)


	def chk000(self, state=None):
		'''Scale Joint
		'''
		self.sb.toggleWidgets(setUnChecked='chk001-2')
		# self.sb.rigging.tb000.ctxMenu.s000.setValue(pm.jointDisplayScale(query=1)) #init global joint display size


	def chk001(self, state=None):
		'''Scale IK
		'''
		self.sb.toggleWidgets(setUnChecked='chk000, chk002')
		# self.sb.rigging.tb000.ctxMenu.setValue(pm.ikHandleDisplayScale(query=1)) #init IK handle display size
		

	def chk002(self, state=None):
		'''Scale IK/FK
		'''
		self.sb.toggleWidgets(setUnChecked='chk000-1')
		# self.sb.rigging.tb000.ctxMenu.setValue(pm.jointDisplayScale(query=1, ikfk=1)) #init IKFK display size


	def s000(self, value=None):
		'''Scale Joint/IK/FK
		'''
		value = self.sb.rigging.tb000.ctxMenu.value()

		# if self.sb.rigging.chk000.isChecked():
		# 	pm.jointDisplayScale(value) #set global joint display size
		# elif self.sb.rigging.chk001.isChecked():
		# 	pm.ikHandleDisplayScale(value) #set global IK handle display size
		# else: #self.sb.rigging.chk002.isChecked():
		# 	pm.jointDisplayScale(value, ikfk=1) #set global IKFK display size


	def tb000(self, state=None):
		'''Toggle Display Local Rotation Axes
		'''
		tb = self.sb.rigging.tb000

		# joints = pm.ls(type="joint") #get all scene joints

		# state = pm.toggle(joints[0], query=1, localAxis=1)
		# if tb.ctxMenu.isChecked():
		# 	if not state:
		# 		toggle=True
		# else:
		# 	if state:
		# 		toggle=True

		# if toggle:
		# 	pm.toggle(joints, localAxis=1) #set display off

		# self.sb.messageBox('Display Local Rotation Axes:<hl>'+str(state)+'</hl>')


	def tb001(self, state=None):
		'''Orient Joints
		'''
		tb = self.sb.rigging.tb001

		# orientJoint = 'xyz' #orient joints
		# if tb.ctxMenu.isChecked():
		# 	orientJoint = 'none' #orient joint to world

		# pm.joint(edit=1, orientJoint=orientJoint, zeroScaleOrient=1, ch=1)


	def tb002(self, state=None):
		'''Constraint: Parent
		'''
		tb = self.sb.rigging.tb002

		template = tb.ctxMenu.chk004.isChecked()

		objects = list(Slots_max.bitArrayToArray(rt.selection))

		for obj in objects[:-1]:
			obj.parent = objects[-1]

			if template:
				obj.isFrozen = True


	def tb003(self, state=None):
		'''Create Locator at Selection
		'''
		tb = self.sb.rigging.tb003

		suffix = tb.ctxMenu.t000.text()
		stripDigits = tb.ctxMenu.chk005.isChecked()
		strip = tb.ctxMenu.t001.text()
		parent = tb.ctxMenu.chk006.isChecked()
		scale = tb.ctxMenu.s001.value()
		lockTranslate = tb.ctxMenu.chk007.isChecked()
		lockRotation = tb.ctxMenu.chk008.isChecked()
		lockScale = tb.ctxMenu.chk009.isChecked()
		freezeTransforms = tb.ctxMenu.chk010.isChecked()
		bakeChildPivot = tb.ctxMenu.chk011.isChecked()
		remove = tb.ctxMenu.chk015.isChecked()

		selection = pm.ls(selection=True)
		Rigging.createLocatorAtObject(selection, suffix=suffix, stripDigits=stripDigits, strip=strip, scale=scale, parent=parent, bakeChildPivot=bakeChildPivot, 
			freezeTransforms=freezeTransforms, lockTranslate=lockTranslate, lockRotation=lockRotation, lockScale=lockScale, remove=remove)


	def tb004(self, state=None):
		'''Lock/Unlock Attributes
		'''
		tb = self.sb.rigging.tb004

		lockTranslate = tb.ctxMenu.chk012.isChecked()
		lockRotation = tb.ctxMenu.chk013.isChecked()
		lockScale = tb.ctxMenu.chk014.isChecked()

		sel = pm.ls(selection=True, transforms=1, long=True)
		for obj in sel:

			attrs_and_state = {
				('tx','ty','tz'):lockTranslate, #attributes and state. ex. ('tx','ty','tz'):False
				('rx','ry','rz'):lockRotation, 
				('sx','sy','sz'):lockScale
			}

			[[pm.setAttr('{}.{}'.format(obj, i), lock=v) for i in k] for k, v in attrs_and_state.items()]


	@Slots.hideMain
	def b000(self):
		'''Object Transform Limit Attributes
		'''
		node = rt.selection
		if not node:
			self.sb.messageBox('Operation requires a single selected object.')
			return

		'finish converting from Maya version.  Slots_max.getParameterValuesMax is not likely doable in the same sense getParameterValuesMEL was'
		params = ['enableTranslationX','translationX','enableTranslationY','translationY','enableTranslationZ','translationZ',
			'enableRotationX','rotationX','enableRotationY','rotationY','enableRotationZ','rotationZ',
			'enableScaleX','scaleX','enableScaleY','scaleY','enableScaleZ','scaleZ']

		attrs = Slots_max.getParameterValuesMax(node, 'transformLimits', params)
		self.setAttributeWindow(node, fn=Slots_max.setParameterValuesMax, 'transformLimits', **attrs)


	def b001(self):
		'''Connect Joints
		'''
		pm.connectJoint(cm=1)


	def b002(self):
		'''Insert Joint Tool
		'''
		pm.setToolTo('insertJointContext') #insert joint tool


	def b004(self):
		'''Reroot
		'''
		pm.reroot() #re-root joints


	def b006(self):
		'''Constraint: Point
		'''
		pm.pointConstraint(offset=[0,0,0], weight=1)


	def b007(self):
		'''Constraint: Scale
		'''
		pm.scaleConstraint(offset=[1,1,1], weight=1)


	def b008(self):
		'''Constraint: Orient
		'''
		pm.orientConstraint(offset=[0,0,0], weight=1)


	def b009(self):
		'''Constraint: Aim
		'''
		pm.aimConstraint(offset=[0,0,0], weight=1, aimVector=[1,0,0], upVector=[0,1,0], worldUpType="vector", worldUpVector=[0,1,0])


	def b010(self):
		'''Constraint: Pole Vector
		'''
		pm.orientConstraint(offset=[0,0,0], weight=1)


	@staticmethod
	def createLocatorAtObject(objects, suffix='_LOC', stripDigits=False, strip='', scale=1, parent=False, freezeTransforms=True, 
					bakeChildPivot=True, lockTranslate=False, lockRotation=False, lockScale=False, remove=False, _fullPath=False):
		'''Create locators with the same transforms as any selected object(s).
		If there are vertices selected it will create a locator at the center of the selected vertices bounding box.

		Parameters:
			objects (str)(list): A list of objects, or an object name to create locators at.
			suffix (str): A string appended to the end of the created locators name. (default: '_LOC') '_LOC#'
			stripDigits (bool): Strip numeric characters from the string. If the resulting name is not unique, maya will append a trailing digit. (default=False)
			strip (str): Strip a specific character set from the locator name. The locators name is based off of the selected objects name. (default=None)
			scale (float) = The scale of the locator. (default=1)
			parent (bool): Parent to object to the locator. (default=False)
			freezeTransforms (bool): Freeze transforms on the locator. (default=True)
			bakeChildPivot (bool): Bake pivot positions on the child object. (default=True)
			lockTranslate (bool): Lock the translate values of the child object. (default=False)
			lockRotation (bool): Lock the rotation values of the child object. (default=False)
			lockScale (bool): Lock the scale values of the child object. (default=False)
			remove (bool): Removes the locator and any child locks. (not valid with component selections) (default=False)
			_fullPath (bool): Internal use only (recursion). Use full path names for Dag objects. This can prevent naming conflicts when creating the locator. (default=False)

		Example: createLocatorAtSelection(strip='_GEO', suffix='', stripDigits=True, parent=True, lockTranslate=True, lockRotation=True)
		'''
		def _formatName(name, stripDigits=stripDigits, strip=strip, suffix=suffix):
			if stripDigits:
				name_ = ''.join([i for i in name if not i.isdigit()])	
			return name_.replace(strip, '')+suffix

		def _create_locator(obj, objName, stripDigits=stripDigits, strip=strip, suffix=suffix, scale=scale, _fullPath=_fullPath):
			locName = _formatName(objName, stripDigits, strip, suffix)

			loc = pm.spaceLocator(name=locName)
			if not any([loc, _fullPath]): #if locator creation fails; try again using the objects fullpath name.
				_retry_using_fullPath()

			pm.scale(loc, scale, scale, scale) #scale the locator
			return loc

		def _parent(obj, loc, freezeTransforms=freezeTransforms, bakeChildPivot=bakeChildPivot, state=1):
			objParent = pm.listRelatives(obj, parent=1)
			pm.parent(obj, loc)
			pm.parent(loc, objParent)

			if freezeTransforms:
				_set_lock_state(obj, translate=1, rotate=1, scale=1, state=0) #assure attributes are unlocked.
				pm.makeIdentity(loc, apply=True, normal=1) #1=the normals on polygonal objects will be frozen. 2=the normals on polygonal objects will be frozen only if its a non-rigid transformation matrix.

			if bakeChildPivot:
				pm.select(obj); pm.mel.BakeCustomPivot() #bake pivot on child object. Requires a select

		def _set_lock_state(obj, translate=lockTranslate, rotate=lockRotation, scale=lockScale, state=1):
			if isinstance(obj, str):
				obj = pm.ls(obj.split('|')[-1], transforms=1)[0]

			if type_(obj)=='locator':
				obj = pm.listRelatives(obj, children=1, type='transform')[0]

			if translate: #lock translation values
				[pm.setAttr(getattr(obj, attr), lock=state) for attr in ('tx','ty','tz')] #pm.setAttr(obj[0].translate, lock=state)
			if rotate: #lock rotation values
				[pm.setAttr(getattr(obj, attr), lock=state) for attr in ('rx','ry','rz')]
			if scale: #lock scale values
				[pm.setAttr(getattr(obj, attr), lock=state) for attr in ('sx','sy','sz')]

		def _try_parent_and_lock(obj, loc, objName, parent=parent, _fullPath=_fullPath):
			try:
				if parent: #parent
					_parent(obj, loc)
				_set_lock_state(obj)
			except Exception as error:
				try:
					if not any([loc, _fullPath]): #try again using the objects full path name.
						_retry_using_fullPath()
				except:
					print ('# Error: {}: {} #'.format(objName, error))
					pm.delete(loc)

		def _remove_locators(obj):
			if not type_(obj) == 'locator':
				obj = pm.listRelatives(obj, parent=1)
				print (obj, type_(obj))
				if not obj or not type_(obj[0]) == 'locator':
					return False
			pm.ungroup(obj)
			pm.delete(obj)

		type_ = lambda obj: pm.listRelatives(obj, shapes=1)[0].type()

		_retry_using_fullPath = lambda: Rigging.createLocatorAtObject(suffix=suffix, stripDigits=stripDigits, strip=strip, 
			parent=parent, scale=scale, lockTranslate=lockTranslate, lockRotation=lockRotation, lockScale=lockScale, _fullPath=1)

		# pm.undoInfo(openChunk=1)
		objects = pm.ls(objects, long=_fullPath, objectsOnly=True)

		sel_verts = pm.filterExpand(objects, sm=31) #returns a string list.
		if sel_verts: #vertex selection

			objName = sel_verts[0].split('.')[0]
			obj = pm.ls(objName)
			
			loc = _create_locator(obj, objName)

			xmin, ymin, zmin, xmax, ymax, zmax = pm.exactWorldBoundingBox(sel_verts)
			x, y, z = pos = ((xmin + xmax) / 2, (ymin + ymax) / 2, (zmin + zmax) / 2)
			pm.move(x, y, z, loc)

			_try_parent_and_lock(obj, loc, objName)

		else: #object selection
			for obj in objects:

				if remove:
					if pm.objExists(obj): #if the locator hasnt already been deleted by another child.
						_set_lock_state(obj, translate=1, rotate=1, scale=1, state=0)
						_remove_locators(obj)
					continue

				objName = obj.name()

				loc = _create_locator(obj, objName)

				tempConst = pm.parentConstraint(obj, loc, mo=False)
				pm.delete(tempConst)

				_try_parent_and_lock(obj, loc, objName)
		# pm.undoInfo(closeChunk=1)








#module name
print (__name__)
# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------


import os, sys, time
from PySide2 import QtCore, QtWidgets, QtGui


class WorkIndicator(QtWidgets.QDialog):
	def __init__(self, parent=None):
		super().__init__(parent)

		self.setAttribute(QtCore.Qt.WA_TranslucentBackground)
		self.setWindowFlags(QtCore.Qt.FramelessWindowHint | QtCore.Qt.WindowTitleHint | QtCore.Qt.WindowStaysOnTopHint)
		self.setModal(True)

		gif_path = "O:/Cloud/Code/_scripts/uitk/uitk/widgets/loading_indicator.gif"
		self.movie = QtGui.QMovie(gif_path, parent=self)
		self.movie.setCacheMode(QtGui.QMovie.CacheAll)
		self.movie.setScaledSize(QtCore.QSize(75, 75))

		label = QtWidgets.QLabel(self)
		label.setMovie(self.movie)
		label.setAlignment(QtCore.Qt.AlignCenter)

		layout = QtWidgets.QVBoxLayout(self)
		layout.addWidget(label)
		self.setLayout(layout)

		self.setFixedSize(self.movie.frameRect().size())

	def start(self):
		self.movie.start()
		self.show()
		self.move(QtGui.QCursor.pos() - QtCore.QPoint(self.width()/2, self.height()/2))

	def stop(self):
		self.movie.stop()
		self.accept()


class Task(QtCore.QThread):
	complete = QtCore.Signal()

	def __init__(self):
		super().__init__()
		self.work_indicator = WorkIndicator()
		self.complete.connect(self.work_indicator.stop)

	def start_task(self, task_function):
		self.task_function = task_function
		self.finished.connect(self.on_finished)
		self.start()
		self.work_indicator.start()

	def run(self):
		self.task_function()
		self.complete.emit()

	def on_finished(self):
		self.finished.disconnect(self.on_finished)
		self.task_function = None
		self.deleteLater()

# --------------------------------------------------------------------------------------------








# --------------------------------------------------------------------------------------------
if __name__=='__main__':
	class MainWindow(QtWidgets.QWidget):
		def __init__(self):
			super().__init__()

			layout = QtWidgets.QVBoxLayout()
			self.setLayout(layout)

			button = QtWidgets.QPushButton("Perform Long Running Task")
			button.clicked.connect(self.perform_long_running_task)
			layout.addWidget(button)

			self.task = Task()

		def perform_long_running_task(self):
			# Create a separate thread to perform the long running task.
			self.task.start_task(self.long_running_task)
			

		def long_running_task(self):
			time.sleep(2)  # Move the time.sleep(5) to this method

	app = QtWidgets.QApplication([])
	main = MainWindow()
	main.show()
	sys.exit(app.exec_())

# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------



# --------------------------------------------------------------------------------------------
# deprecated:
# --------------------------------------------------------------------------------------------


# !/usr/bin/python
# coding=utf-8
from PySide2 import QtCore, QtGui, QtWidgets


class Attributes(object):
	'''Methods for setting widget Attributes.
	'''

	def setAttributes(self, obj=None, **kwargs):
		'''Works with attributes passed in as a dict or kwargs.
		If attributes are passed in as a dict, kwargs are ignored.

		Parameters:
			obj (obj): the child obj or widgetAction to set attributes for. (default=self)
			**kwargs = The keyword arguments to set.
		'''
		if not kwargs: #if no attributes given.
			return

		if not obj:
			obj = self

		for attr, value in kwargs.items():
			try:
				getattr(obj, attr)(value)

			except:
				#print ('setAttributes:', attr, value)
				self.setCustomAttribute(obj, attr, value)


	def setCustomAttribute(self, w, attr, value):
		'''Attributes that throw an AttributeError in 'setAttributes' are sent here, where they can be assigned a value.
		Custom attributes can be set using a trailing underscore convention to aid readability, and differentiate them from standard attributes.

		Parameters:
			w (obj): The child widget or widgetAction to set attributes for.
			attr (str): Custom keyword attribute.
			value (str): The value corresponding to the given attr.

		attributes:
			copy_ (obj): The widget to copy attributes from.
			setSize_ (list): The size as an x and y value. ie. (40, 80)
			setWidth_ (int): The desired width.
			setHeight_ (int): The desired height.
			setPosition_ (QPoint)(str): Move to the given global position and center. valid: <QPoint>, 'cursor', 
			addMenu_ (QMenu) = Used for adding additional menus to a parent menu. ex. parentMenu = Menu(); childMenu = Menu('Create', addMenu_=parentMenu)
			insertSeparator_ (bool): Insert a line separater before the new widget.
			setLayoutDirection_ (str): Set the layout direction using a string value. ie. 'LeftToRight'
			setAlignment_ (str): Set the alignment using a string value. ie. 'AlignVCenter'
			setButtonSymbols_ (str): Set button symbols using a string value. ex. ie. 'PlusMinus'
			setMinMax_ (str): Set the min, max, and step value using a string value. ex. '.01-10 step.1'
			setCheckState_ (int): Set a tri-state checkbox state using an integer value. 0(unchecked), 1(partially checked), 2(checked).
		'''
		if attr=='copy_':
			w.setObjectName(value.objectName())
			w.resize(value.size())
			w.setText(value.text())
			w.setWhatsThis(value.whatsThis())

		elif attr=='setSize_':
			x, y = value
			w.resize(QtCore.QSize(x, y))

		elif attr=='setWidth_':
			w.setFixedWidth(value)
			# w.resize(value, w.size().height())

		elif attr=='setHeight_':
			w.setFixedHeight(value)
			# w.resize(w.size().width(), value)

		elif attr=='setPosition_':
			if value=='cursor':
				value = QtGui.QCursor.pos()
			w.move(w.mapFromGlobal(value - w.rect().center())) #move and center

		elif attr=='addMenu_':
			value.addMenu(w)

		elif attr=='insertSeparator_':
			if w.__class__.__name__=='QAction':
				self.insertSeparator(w)

		elif attr=='setLayoutDirection_':
			self.setAttributes(w, setLayoutDirection=getattr(QtCore.Qt, value))

		elif attr=='setAlignment_':
			self.setAttributes(w, setAlignment=getattr(QtCore.Qt, value))

		elif attr=='setButtonSymbols_':
			self.setAttributes(w, setButtonSymbols=getattr(QtWidgets.QAbstractSpinBox, value))

		#presets
		elif attr=='setMinMax_':
			self.setMinMax(w, value)

		elif attr=='setSpinBoxByValue_':
			self.setSpinBoxByValue(w, value)

		elif attr=='setCheckState_':
			state = {0:QtCore.Qt.CheckState.Unchecked, 1:QtCore.Qt.CheckState.PartiallyChecked, 2:QtCore.Qt.CheckState.Checked}
			w.setCheckState(state[value])

		else:
			print('Error: {} has no attribute {}'.format(w, attr))


	def setMinMax(self, spinbox, value):
		'''Set the minimum, maximum, and step values for a spinbox using a shorthand string value.

		Parameters:
			spinbox (obj): spinbox widget.
			value (str): value as shorthand string. ie. '0.00-100 step1'
		'''
		stepStr = value.split()[1].strip('step')
		step = float(stepStr)
		decimals = len(stepStr.split('.')[-1])

		spanStr = value.split('-')
		minimum = float(spanStr[0])
		maximum = float(spanStr[1].split()[0])

		if hasattr(spinbox, 'setDecimals'):
			self.setAttributes(spinbox, setDecimals=decimals)

		self.setAttributes(spinbox, setMinimum=minimum, setMaximum=maximum, setSingleStep=step, setButtonSymbols_='NoButtons')


	def setSpinBoxByValue(self, spinbox, value):
		'''Set a spinbox's attributes according to a given value.

		Parameters:
			spinbox (obj): spinbox widget.
			value (multi) = attribute value.
		'''
		maximum = spinbox.maximum()
		minimum = -maximum

		if isinstance(value, (int, bool)):
			step = spinbox.singleStep()

			if isinstance(value, bool):
				value = int(value)
				minimum = 0
				maximum = 1

			self.setAttributes(spinbox, setValue=value, setMinimum=minimum, setMaximum=maximum, 
				setSingleStep=step, setButtonSymbols_='NoButtons',)

		elif isinstance(value, float):
			decimals = str(value)[::-1].find('.') #get decimal places
			step = Attributes.moveDecimalPoint(1, -decimals)

			self.setAttributes(spinbox, setValue=value, setMinimum=minimum, setMaximum=maximum, 
				setSingleStep=step, setDecimals=decimals, setButtonSymbols_='NoButtons',)


	@staticmethod
	def moveDecimalPoint(num, decimal_places):
		'''Move the decimal place in a given number.

		Parameters:
			decimal_places (int): decimal places to move. (works only with values 0 and below.)
		
		Return:
			(float) the given number with it's decimal place moved by the desired amount.
		
		ex. moveDecimalPoint(11.05, -2) Return: 0.1105
		'''
		for _ in range(abs(decimal_places)):

			if decimal_places>0:
				num *= 10; #shifts decimal place right
			else:
				num /= 10.; #shifts decimal place left

		return float(num)









if __name__ == "__main__":
	import sys
	app = QtWidgets.QApplication.instance() or QtWidgets.QApplication(sys.argv) #return the existing QApplication object, or create a new one if none exists.

	sys.exit(app.exec_())



# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------

'''
Promoting a widget in designer to use a custom class:
>	In Qt Designer, select all the widgets you want to replace, 
		then right-click them and select 'Promote to...'. 

>	In the dialog:
		Base Class:		Class from which you inherit. ie. QWidget
		Promoted Class:	Name of the class. ie. "MyWidget"
		Header File:	Path of the file (changing the extension .py to .h)  ie. myfolder.mymodule.mywidget.h

>	Then click "Add", "Promote", 
		and you will see the class change from "QWidget" to "MyWidget" in the Object Inspector pane.
'''

# depricated ------------------------------------------------------------------------







# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------


# !/usr/bin/python
# coding=utf-8
from uitk.slots.blender import *
from uitk.slots.scene import Scene



class Scene_blender(Scene, Slots_blender):
	def __init__(self, *args, **kwargs):
		Slots_blender.__init__(self, *args, **kwargs)
		Scene.__init__(self, *args, **kwargs)

		cmb = self.sb.scene.draggable_header.ctxMenu.cmb000
		items = ['Node Editor', 'Outlinder', 'Content Browser', 'Optimize Scene Size', 'Prefix Hierarchy Names', 'Search and Replace Names']
		cmb.addItems_(items, 'Maya Scene Editors')


	def cmb000(self, index=-1):
		'''Editors
		'''
		cmb = self.sb.scene.draggable_header.ctxMenu.cmb000

		if index>0:
			text = cmb.items[index]
			if text=='Node Editor':
				mel.eval('NodeEditorWindow;') #
			elif text=='Outlinder':
				mel.eval('OutlinerWindow;') #
			elif text=='Content Browser':
				mel.eval('ContentBrowserWindow;') #
			elif text=='Optimize Scene Size':
				mel.eval('cleanUpScene 2;')
			elif text=='Prefix Hierarchy Names':
				mel.eval('prefixHierarchy;') #Add a prefix to all hierarchy names.
			elif text=='Search and Replace Names':
				mel.eval('SearchAndReplaceNames;') #performSearchReplaceNames 1; #Rename objects in the scene.
			cmb.setCurrentIndex(0)


	def tb000(self, state=None):
		'''Convert Case
		'''
		tb = self.sb.scene.tb000

		case = tb.ctxMenu.cmb001.currentText()

		selection = pm.ls(sl=1)
		objects = selection if selection else pm.ls(objectsOnly=1)
		self.setCase(objects, case)


	@Slots_blender.undoChunk
	def rename(self, frm, to, objects=[], regEx=False, ignoreCase=False):
		'''Rename scene objects.

		Parameters:
			frm (str): Current name. An asterisk denotes startswith*, *endswith, *contains*, and multiple search strings can be separated by pipe ('|') chars.
				frm - Search exact.
				*frm* - Search contains chars.
				*frm - Search endswith chars.
				frm* - Search startswith chars.
				frm|frm - Search any of.  can be used in conjuction with other modifiers.
			to (str): Desired name: An optional asterisk modifier can be used for formatting
				to - replace all.
				*to* - replace only.
				*to - replace suffix.
				**to - append suffix.
				to* - replace prefix.
				to** - append prefix.
			objects (list): The objects to rename. If nothing is given, all scene objects will be renamed.
			regEx (bool): If True, regular expression syntax is used instead of the default '*' and '|' modifiers.
			ignoreCase (bool): Ignore case when searching. Applies only to the 'frm' parameter's search.

		ex. rename(r'Cube', '*001', regEx=True) #replace chars after frm on any object with a name that contains 'Cube'. ie. 'polyCube001' from 'polyCube'
		ex. rename(r'Cube', '**001', regEx=True) #append chars on any object with a name that contains 'Cube'. ie. 'polyCube1001' from 'polyCube1'
		'''
		# pm.undoInfo (openChunk=1)
		objects = pm.ls(objectsOnly=1) if not objects else objects
		names = self.findStrAndFormat([obj.name() for obj in objects], to, frm, regEx=regEx, ignoreCase=ignoreCase, returnOldNames=True)
		print ('# Rename: Found {} matches. #'.format(len(names)))

		for oldName, newName in names:
			try:
				if pm.objExists(oldName):
					n = pm.rename(oldName, newName) #Rename the object with the new name
					if not n==newName:
						print ('# Warning: Attempt to rename "{}" to "{}" failed. Renamed instead to "{}". #'.format(oldName, newName, n))
					else:
						print ('# Result: Successfully renamed "{}" to "{}". #'.format(oldName, newName))

			except Exception as e:
				if not pm.ls(oldName, readOnly=True)==[]: #ignore read-only errors.
					print ('# Error: Attempt to rename "{}" to "{}" failed. {} #'.format(oldName, newName, str(e).rstrip()))
		# pm.undoInfo (closeChunk=1)


	@Slots_blender.undoChunk
	def setCase(self, objects=[], case='caplitalize'):
		'''Rename objects following the given case.

		Parameters:
			objects (str)(list): The objects to rename. default:all scene objects
			case (str): Desired case using python case operators. 
				valid: 'upper', 'lower', 'caplitalize', 'swapcase' 'title'. default:'caplitalize'

		Example: setCase(pm.ls(sl=1), 'upper')
		'''
		# pm.undoInfo(openChunk=1)
		for obj in pm.ls(objects):
			name = obj.name()

			newName = getattr(name, case)()
			try:
				pm.rename(name, newName)
			except Exception as error:
				if not pm.ls(obj, readOnly=True)==[]: #ignore read-only errors.
					print (name+': ', error)
		# pm.undoInfo(closeChunk=1)









#module name
print (__name__)
# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------


# !/usr/bin/python
# coding=utf-8
from uitk.slots.blender import *
from uitk.slots.polygons import Polygons



class Polygons_blender(Polygons, Slots_blender):
	def __init__(self, *args, **kwargs):
		Slots_blender.__init__(self, *args, **kwargs)
		Polygons.__init__(self, *args, **kwargs)

		cmb000 = self.sb.polygons.draggable_header.ctxMenu.cmb000
		items = ['Extrude','Bevel','Bridge','Combine','Merge Vertex','Offset Edgeloop','Edit Edgeflow','Extract Curve','Poke','Wedge','Assign Invisible']
		cmb000.addItems_(items, 'Polygon Editors')


	def cmb000(self, index=-1):
		'''Editors
		'''
		cmb = self.sb.polygons.draggable_header.ctxMenu.cmb000

		if index>0:
			text = cmb.items[index]
			if text=='Extrude':
				pm.mel.PolyExtrudeOptions()
			elif text=='Bevel':
				pm.mel.BevelPolygonOptions()
			elif text=='Bridge':
				pm.mel.BridgeEdgeOptions()
			elif text=='Combine':
				pm.mel.CombinePolygonsOptions()
			elif text=='Merge Vertex':
				pm.mel.PolyMergeOptions()
			elif text=='Offset Edgeloop':
				pm.mel.DuplicateEdgesOptions()
			elif text=='Edit Edgeflow':
				pm.mel.PolyEditEdgeFlowOptions()
			elif text=='Extract Curve':
				pm.mel.CreateCurveFromPolyOptions()
			elif text=='Poke':
				pm.mel.PokePolygonOptions()
			elif text=='Wedge':
				pm.mel.WedgePolygonOptions()
			elif text=='Assign Invisible':
				pm.mel.PolyAssignSubdivHoleOptions()
			cmb.setCurrentIndex(0)


	def tb000(self, state=None):
		'''Merge Vertices
		'''
		tb = self.sb.polygons.tb000

		tolerance = float(tb.ctxMenu.s002.value())
		objects = pm.ls(selection=1, objectsOnly=1, flatten=1)
		componentMode = pm.selectMode(query=1, component=1)

		if not objects:
			return 'Error: <strong>Nothing selected</strong>.<br>Operation requires an object or vertex selection.'

		for obj in objects:
			object_vert_sel = pm.ls(obj, sl=1)
			if componentMode: #merge selected components.
				if pm.filterExpand(selectionMask=31): #selectionMask=vertices
					pm.polyMergeVertex(object_vert_sel, distance=tolerance, alwaysMergeTwoVertices=True, constructionHistory=True)
				else: #if selection type =edges or facets:
					pm.mel.MergeToCenter()

			else: #if object mode. merge all vertices on the selected object.
				for n, vert in enumerate(object_vert_sel):
					if not self.sb.polygons.progressBar.step(n, len(object_vert_sel)): #register progress while checking for cancellation:
						break

					# get number of vertices
					count = pm.polyEvaluate(obj, vertex=1)
					vertices = str(obj) + ".vtx [0:" + str(count) + "]" # mel expression: select -r geometry.vtx[0:1135];
					pm.polyMergeVertex(vertices, distance=tolerance, alwaysMergeTwoVertices=False, constructionHistory=False)

				#return to original state
				pm.select(clear=1)
				pm.select(objects)


	@Slots_blender.attr
	def tb001(self, state=None):
		'''Bridge
		'''
		tb = self.sb.polygons.tb001

		divisions = tb.ctxMenu.s003.value()

		selection = pm.ls(sl=1)
		edges = pm.filterExpand(selection, selectionMask=32, expand=1) #get edges from selection

		node = pm.polyBridgeEdge(edges, divisions=divisions) #bridge edges
		pm.polyCloseBorder(edges) #fill edges if they lie on a border
		return node


	def tb002(self, state=None):
		'''Combine
		'''
		tb = self.sb.polygons.tb002

		# pm.polyUnite( 'plg1', 'plg2', 'plg3', name='result' ) #for future reference. if more functionality is needed use polyUnite
		if tb.ctxMenu.chk000.isChecked():
			sel = pm.ls(sl=1, objectsOnly=1)
			if not sel:
				return 'Error: <strong>Nothing selected</strong>.'
			objName = sel[0].name()
			objParent = pm.listRelatives(objName, parent=1)
			#combine
			newObj = pm.polyUnite(ch=1, mergeUVSets=1, centerPivot=1)
			#rename using the first selected object
			pm.bakePartialHistory(objName, all=True)
			objName_ = pm.rename(newObj[0], objName)
			#reparent
			pm.parent(objName_, objParent)
		else:
			pm.mel.CombinePolygons()


	@Slots_blender.attr
	def tb003(self, state=None):
		'''Extrude
		'''
		tb = self.sb.polygons.tb003

		keepFacesTogether = tb.ctxMenu.chk002.isChecked() #keep faces/edges together.
		divisions = tb.ctxMenu.s004.value()

		selection = pm.ls(sl=1)
		if pm.selectType(query=1, facet=1): #face selection
			pm.polyExtrudeFacet(edit=1, keepFacesTogether=keepFacesTogether, divisions=divisions)
			mel.eval('PolyExtrude;')
			# return pm.polyExtrudeFacet(selection, ch=1, keepFacesTogether=keepFacesTogether, divisions=divisions)

		elif pm.selectType(query=1, edge=1): #edge selection
			pm.polyExtrudeEdge(edit=1, keepFacesTogether=keepFacesTogether, divisions=divisions)
			mel.eval('PolyExtrude;')
			# return pm.polyExtrudeEdge(selection, ch=1, keepFacesTogether=keepFacesTogether, divisions=divisions)

		elif pm.selectType(query=1, vertex=1): #vertex selection
			pm.polyExtrudeVertex(edit=1, width=0.5, length=1, divisions=divisions)
			mel.eval('PolyExtrude;')
			# return polyExtrudeVertex(selection, ch=1, width=0.5, length=1, divisions=divisions)


	@Slots_blender.attr
	def tb004(self, state=None):
		'''Bevel (Chamfer)
		'''
		tb = self.sb.polygons.tb004

		width = float(tb.ctxMenu.s000.value())
		chamfer = True
		segments = 1

		return pm.polyBevel3(fraction=width, offsetAsFraction=1, autoFit=1, depth=1, mitering=0, 
			miterAlong=0, chamfer=chamfer, segments=segments, worldSpace=1, smoothingAngle=30, subdivideNgons=1,
			mergeVertices=1, mergeVertexTolerance=0.0001, miteringAngle=180, angleTolerance=180, ch=0)


	def tb005(self, state=None):
		'''Detach
		'''
		tb = self.sb.polygons.tb005

		duplicate = tb.ctxMenu.chk014.isChecked()
		separate = tb.ctxMenu.chk015.isChecked()

		vertexMask = pm.selectType (query=True, vertex=True)
		edgeMask = pm.selectType (query=True, edge=True)
		facetMask = pm.selectType (query=True, facet=True)

		component_sel = pm.ls(sl=1)
		if not component_sel:
			return 'Error: <strong>Nothing selected</strong>.'

		if vertexMask:
			pm.mel.polySplitVertex()

		elif facetMask:
			extract = pm.polyChipOff(component_sel, ch=1, keepFacesTogether=1, dup=duplicate, off=0)
			if separate:
				try:
					splitObjects = pm.polySeparate(component_sel)
				except:
					splitObjects = pm.polySeparate(pm.ls(component_sel, objectsOnly=1))
			pm.select(splitObjects[-1])
			return extract

		else:
			pm.mel.DetachComponent()


	@Slots_blender.attr
	def tb006(self, state=None):
		'''Inset Face Region
		'''
		tb = self.sb.polygons.tb006

		selected_faces = pm.polyEvaluate(faceComponent=1)
		if isinstance(selected_faces, str): #'Nothing counted : no polygonal object is selected.'
			return 'Error: <strong>Nothing selected</strong>.<br>Operation requires a face selection.'

		offset = float(tb.ctxMenu.s001.value())
		return pm.polyExtrudeFacet(selected_faces, keepFacesTogether=1, pvx=0, pvy=40.55638003, pvz=33.53797107, divisions=1, twist=0, taper=1, offset=offset, thickness=0, smoothingAngle=30)


	def tb007(self, state=None):
		'''Divide Facet
		'''
		tb = self.sb.polygons.tb007

		dv=u=v=0
		if tb.ctxMenu.chk008.isChecked(): #Split U
			u=2
		if tb.ctxMenu.chk009.isChecked(): #Split V
			v=2

		mode = 0 #The subdivision mode. 0=quads, 1=triangles
		subdMethod = 1 #subdivision type: 0=exponential(traditional subdivision) 1=linear(number of faces per edge grows linearly)
		if tb.ctxMenu.chk010.isChecked(): #tris
			mode=dv=1
			subdMethod=0
		if all([tb.ctxMenu.chk008.isChecked(), tb.ctxMenu.chk009.isChecked()]): #subdivide once into quads
			dv=1
			subdMethod=0
			u=v=0
		#perform operation
		selectedFaces = pm.filterExpand (pm.ls(sl=1), selectionMask=34, expand=1)
		if selectedFaces:
			for face in selectedFaces: #when performing polySubdivideFacet on multiple faces, adjacent subdivided faces will make the next face an n-gon and therefore not able to be subdivided. 
				pm.polySubdivideFacet(face, divisions=0, divisionsU=2, divisionsV=2, mode=0, subdMethod=1)
		else:
			return 'Error: <strong>Nothing selected</strong>.<br>Operation requires a face selection.'


	def tb008(self, state=None):
		'''Boolean Operation
		'''
		tb = self.sb.polygons.tb008

		if tb.ctxMenu.chk011.isChecked(): #union
			pm.mel.PolygonBooleanIntersection()

		if tb.ctxMenu.chk012.isChecked(): #difference
			pm.mel.PolygonBooleanDifference()

		if tb.ctxMenu.chk013.isChecked(): #intersection
			pm.mel.PolygonBooleanIntersection()


	def tb009(self, state=None):
		'''Snap Closest Verts
		'''
		tb = self.sb.polygons.tb009

		tolerance = tb.ctxMenu.s005.value()
		freezetransforms = tb.ctxMenu.chk016.isChecked()

		selection = pm.ls(sl=1, objectsOnly=1, type='transform')
		if len(selection)>1:
			obj1, obj2 = selection
			self.snapClosestVerts(obj1, obj2, tolerance, freezetransforms)
		else:
			return 'Error: <strong>Nothing selected</strong>.<br>Operation requires at least two selected objects.'


	@Slots_blender.attr
	def b000(self):
		'''Circularize
		'''
		circularize = pm.polyCircularize(
			constructionHistory=1, 
			alignment=0, 
			radialOffset=0, 
			normalOffset=0, 
			normalOrientation=0, 
			smoothingAngle=30, 
			evenlyDistribute=1, 
			divisions=0, 
			supportingEdges=0, 
			twist=0, 
			relaxInterior=1
		)
		return circularize


	def b001(self):
		'''Fill Holes
		'''
		pm.mel.FillHole()


	def b002(self):
		'''Separate
		'''
		pm.mel.SeparatePolygon()


	def b003(self):
		'''Symmetrize
		'''
		pm.mel.Symmetrize()


	@Slots_blender.attr
	def b004(self):
		'''Slice
		'''
		cuttingDirection = 'Y' #valid values: 'x','y','z' A value of 'x' will cut the object along the YZ plane cutting through the center of the bounding box. 'y':ZX. 'z':XY.

		component_sel = pm.ls(sl=1)
		return pm.polyCut(component_sel, cuttingDirection=cuttingDirection, ch=1)


	def b009(self):
		'''Collapse Component
		'''
		if pm.selectType(query=1, vertex=1):
			pm.mel.MergeToCenter() #merge vertices
		else:
			pm.mel.PolygonCollapse()


	def b012(self):
		'''Multi-Cut Tool
		'''
		pm.mel.dR_multiCutTool()


	def b021(self):
		'''Connect Border Edges
		'''
		pm.mel.performPolyConnectBorders(0)


	def b022(self):
		'''Attach
		'''
		# pm.mel.AttachComponent()
		pm.mel.dR_connectTool()


	def b028(self):
		'''Quad Draw
		'''
		pm.mel.dR_quadDrawTool()


	def b032(self):
		'''Poke
		'''
		pm.mel.PokePolygon()


	def b034(self):
		'''Wedge
		'''
		pm.mel.WedgePolygon()


	def b038(self):
		'''Assign Invisible
		'''
		pm.polyHole(assignHole=1)


	def b043(self):
		'''Target Weld
		'''
		pm.select(deselect=True)
		pm.mel.dR_targetWeldTool()


	def b045(self):
		'''Re-Order Vertices
		'''
		symmetryOn = pm.symmetricModelling(query=True, symmetry=True) #query symmetry state
		if symmetryOn:
			pm.symmetricModelling(symmetry=False)
		pm.mel.setPolygonDisplaySettings("vertIDs") #set vertex id on
		pm.mel.doBakeNonDefHistory(1, "pre") #history must be deleted
		pm.mel.performPolyReorderVertex() #start vertex reorder ctx


	def b046(self):
		'''Split
		'''
		vertexMask = pm.selectType (query=True, vertex=True)
		edgeMask = pm.selectType (query=True, edge=True)
		facetMask = pm.selectType (query=True, facet=True)

		if facetMask:
			pm.mel.performPolyPoke(1)

		elif edgeMask:
			pm.polySubdivideEdge(ws=0, s=0, dv=1, ch=0)

		elif vertexMask:
			pm.mel.polyChamferVtx(0, 0.25, 0)


	def b047(self):
		'''Insert Edgeloop
		'''
		pm.mel.SplitEdgeRingTool()


	def b048(self):
		'''Collapse Edgering
		'''
		pm.mel.bt_polyCollapseEdgeRingTool()


	def b049(self):
		'''Slide Edge Tool
		'''
		pm.mel.SlideEdgeTool()


	def b050(self):
		'''Spin Edge
		'''
		pm.mel.bt_polySpinEdgeTool()


	def b051(self):
		'''Offset Edgeloop
		'''
		pm.mel.performPolyDuplicateEdge(0)


	def b053(self):
		'''Edit Edge Flow
		'''
		pm.polyEditEdgeFlow(adjustEdgeFlow=1)


	@Slots_blender.undoChunk
	def snapClosestVerts(self, obj1, obj2, tolerance=10.0, freezeTransforms=False):
		'''Snap the vertices from object one to the closest verts on object two.

		Parameters:
			obj1 (obj): The object in which the vertices are moved from.
			obj2 (obj): The object in which the vertices are moved to.
			tolerance (float) = Maximum search distance.
			freezeTransforms (bool): Reset the selected transform and all of its children down to the shape level.
		'''
		vertices = mtk.Cmpt.getComponents(obj1, 'vertices')
		closestVerts = self.getClosestVertex(vertices, obj2, tolerance=tolerance, freezeTransforms=freezeTransforms)

		progressBar = mel.eval("$container=$gMainProgressBar");
		pm.progressBar(progressBar, edit=True, beginProgress=True, isInterruptable=True, status="Snapping Vertices ...", maxValue=len(closestVerts)) 

		# pm.undoInfo(openChunk=True)
		for v1, v2 in closestVerts.items():
			if pm.progressBar(progressBar, query=True, isCancelled=True):
				break

			v2Pos = pm.pointPosition(v2, world=True)
			pm.xform(v1, translation=v2Pos, worldSpace=True)

			pm.progressBar(progressBar, edit=True, step=1)
		# pm.undoInfo(closeChunk=True)

		pm.progressBar(progressBar, edit=True, endProgress=True)













#module name
print (__name__)
# --------------------------------------------------------------------------------------------
# Notes
# --------------------------------------------------------------------------------------------


# deprecated:

# 	def tb005(self, state=None):
# 		'''
# 		Detach
# 		'''
# 		tb = self.sb.polygons.tb005
# 		if state=='setMenu':
# 			# tb.ctxMenu.add('QCheckBox', setText='Delete Original', setObjectName='chk007', setChecked=True, setToolTip='Delete original selected faces.')
# 			return

# 		vertexMask = pm.selectType (query=True, vertex=True)
# 		edgeMask = pm.selectType (query=True, edge=True)
# 		facetMask = pm.selectType (query=True, facet=True)

# 		if vertexMask:
# 			mel.eval("polySplitVertex()")

# 		if facetMask:
# 			maskVertex = pm.selectType (query=True, vertex=True)
# 			if maskVertex:
# 				mel.eval("DetachComponent;")
# 			else:
# 				selFace = pm.ls(ni=1, sl=1)
# 				selObj = pm.ls(objectsOnly=1, noIntermediate=1, sl=1) #to errorcheck if more than 1 obj selected

# 				if len(selFace) < 1:
# 					return 'Error: Nothing selected.'

# 				# if len(selObj) > 1:
# 				# 	return 'Error: Only components from a single object can be extracted.'

# 				else:
# 					mel.eval("DetachComponent;")
# 					# pm.undoInfo (openChunk=1)
# 					# sel = str(selFace[0]).split(".") #creates ex. ['polyShape', 'f[553]']
# 					# print(sel)
# 					# extractedObject = "extracted_"+sel[0]
# 					# pm.duplicate (sel[0], name=extractedObject)
# 					# if tb.ctxMenu.chk007.isChecked(): #delete original
# 					# 	pm.delete (selFace)

# 					# allFace = [] #populate a list of all faces in the duplicated object
# 					# numFaces = pm.polyEvaluate(extractedObject, face=1)
# 					# num=0
# 					# for _ in range(numFaces):
# 					# 	allFace.append(extractedObject+".f["+str(num)+"]")
# 					# 	num+=1

# 					# extFace = [] #faces to keep
# 					# for face in selFace:
# 					# 	fNum = str(face.split(".")[0]) #ex. f[4]
# 					# 	extFace.append(extractedObject+"."+fNum)

# 					# delFace = [x for x in allFace if x not in extFace] #all faces not in extFace
# 					# pm.delete (delFace)

# 					# pm.select (extractedObject)
# 					# pm.xform (cpc=1) #center pivot
# 					# pm.undoInfo (closeChunk=1)
# 					# return extractedObject


# !/usr/bin/python
# coding=utf-8
from uitk.slots import Slots



class Rigging(Slots):
	'''
	'''
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)
		'''
		'''
		dh = self.sb.rigging.draggable_header
		dh.ctxMenu.add(self.sb.ComboBox, setObjectName='cmb000', setToolTip='Rigging Editors')

		tb000 = self.sb.rigging.tb000
		tb000.ctxMenu.add('QCheckBox', setText='Joints', setObjectName='chk000', setChecked=True, setToolTip='Display Joints.')
		tb000.ctxMenu.add('QCheckBox', setText='IK', setObjectName='chk001', setChecked=True, setToolTip='Display IK.')
		tb000.ctxMenu.add('QCheckBox', setText='IK\\FK', setObjectName='chk002', setChecked=True, setToolTip='Display IK\\FK.')
		tb000.ctxMenu.add('QDoubleSpinBox', setPrefix='Tolerance: ', setObjectName='s000', setMinMax_='0.00-10 step.5', setValue=1.0, setToolTip='Global Display Scale for the selected type.')
		self.chk000() #init scale joint value

		tb001 = self.sb.rigging.tb001
		tb001.ctxMenu.add('QCheckBox', setText='Align world', setObjectName='chk003', setToolTip='Align joints with the worlds transform.')

		tb002 = self.sb.rigging.tb002
		tb002.ctxMenu.add('QCheckBox', setText='Template Child', setObjectName='chk004', setChecked=False, setToolTip='Template child object(s) after parenting.')		

		tb003 = self.sb.rigging.tb003
		tb003.ctxMenu.add('QDoubleSpinBox', setPrefix='Locator Scale: ', setObjectName='s001', setMinMax_='.000-1000 step1', setValue=1, setToolTip='The scale of the locator.')
		tb003.ctxMenu.add('QLineEdit', setPlaceholderText='Group Suffix:', setText='_GRP', setObjectName='t002', setToolTip='A string appended to the end of the created group\'s name.')
		tb003.ctxMenu.add('QLineEdit', setPlaceholderText='Locator Suffix:', setText='', setObjectName='t000', setToolTip='A string appended to the end of the created locator\'s name.')
		tb003.ctxMenu.add('QLineEdit', setPlaceholderText='Geometry Suffix:', setText='', setObjectName='t001', setToolTip='A string appended to the end of the existing geometry\'s name.')
		tb003.ctxMenu.add('QCheckBox', setText='Strip Suffix', setObjectName='chk016', setToolTip='Strip any of preexisting suffixes from the group name before appending the new ones.\nA suffix is defined as anything trailing an underscore.\nAny user-defined suffixes are stripped by default.')
		tb003.ctxMenu.add('QCheckBox', setText='Strip Digits', setObjectName='chk005', setChecked=True, setToolTip='Strip any trailing numeric characters from the name.\nIf the resulting name is not unique, maya will append a trailing digit.')
		tb003.ctxMenu.add('QCheckBox', setText='Parent', setObjectName='chk006', setChecked=True, setToolTip='Parent to object to the locator.')
		tb003.ctxMenu.add('QCheckBox', setText='Freeze Transforms', setObjectName='chk010', setChecked=True, setToolTip='Freeze transforms on the locator.')
		tb003.ctxMenu.add('QCheckBox', setText='Bake Child Pivot', setObjectName='chk011', setChecked=True, setToolTip='Bake pivot positions on the child object.')
		tb003.ctxMenu.add('QCheckBox', setText='Lock Child Translate', setObjectName='chk007', setChecked=True, setToolTip='Lock the translate values of the child object.')
		tb003.ctxMenu.add('QCheckBox', setText='Lock Child Rotation', setObjectName='chk008', setChecked=True, setToolTip='Lock the rotation values of the child object.')
		tb003.ctxMenu.add('QCheckBox', setText='Lock Child Scale', setObjectName='chk009', setToolTip='Lock the scale values of the child object.')
 
		tb004 = self.sb.rigging.tb004
		tb004.ctxMenu.add('QCheckBox', setText='Translate', setObjectName='chk012', setChecked=False, setToolTip='')
		tb004.ctxMenu.add('QCheckBox', setText='Rotate', setObjectName='chk013', setChecked=False, setToolTip='')
		tb004.ctxMenu.add('QCheckBox', setText='Scale', setObjectName='chk014', setChecked=False, setToolTip='')
		self.sb.connect((tb004.ctxMenu.chk012, tb004.ctxMenu.chk013, tb004.ctxMenu.chk014), 'toggled', 
			[lambda state: self.sb.rigging.tb004.setText('Lock Attributes' 
				if any((tb004.ctxMenu.chk012.isChecked(), tb004.ctxMenu.chk013.isChecked(), tb004.ctxMenu.chk014.isChecked())) else 'Unlock Attributes'), 
			lambda state: self.sb.rigging_submenu.tb004.setText('Lock Transforms' 
				if any((tb004.ctxMenu.chk012.isChecked(), tb004.ctxMenu.chk013.isChecked(), tb004.ctxMenu.chk014.isChecked())) else 'Unlock Attributes')])


	def draggable_header(self, state=None):
		'''Context menu
		'''
		dh = self.sb.rigging.draggable_header


	



# !/usr/bin/python
# coding=utf-8
from PySide2 import QtCore, QtGui, QtWidgets

from uitk.widgets.attributes import Attributes


class Menu(QtWidgets.QMenu, Attributes):
	'''
	Parameters:
		menu_type (str): Menu style. valid parameters are: 'standard', 'context', 'form'
		title (str): Text displayed at the menu's header.
		padding (int): Area surrounding the menu.
		childHeight (int): The minimum height of any child widgets (excluding the 'Apply' button).
		preventHide (bool): Prevent the menu from hiding.
		position (str): Desired menu position. Valid values are: 
			QPoint, tuple as (int, int), 'cursorPos', 'center', 'top', 'bottom', 'right', 'left', 
			'topLeft', 'topRight', 'bottomRight', 'bottomLeft' (Positions relative to parent (requires parent))
	'''
	openMenus = []

	def __init__(self, parent=None, title='', menu_type='standard', position='cursorPos', childHeight=16, preventHide=False, alpha=175, padding=8, **kwargs):
		QtWidgets.QMenu.__init__(self, parent)

		self.menu_type = menu_type
		self.position = position
		self.preventHide = preventHide
		self.childHeight = childHeight
		self.padding = padding
		self.alpha = alpha

		self.setTitle(title)
		self.setWindowFlags(QtCore.Qt.Tool|QtCore.Qt.FramelessWindowHint)
		self.setAttribute(QtCore.Qt.WA_TranslucentBackground)
		self.setSizePolicy(QtWidgets.QSizePolicy.Expanding, QtWidgets.QSizePolicy.Expanding)

		self.childWidgets = set()
		self.layouts = {} #a container for any created layouts.
		self.toggleAllButton = self._addToggleAllButton()
		self.applyButton = self._addApplyButton()

		self.menu_timer = QtCore.QTimer()
		self.menu_timer.timeout.connect(lambda: None if self.rect().contains(QtGui.QCursor.pos()) else self.hide()) #hide the menu if the mouse is not currently over it.

		self.setStyleSheet(f'''
			QMenu {{
				background-color: rgba(50,50,50,{self.alpha});
				padding: {self.padding}px {self.padding}px {self.padding}px {self.padding}px;
			}}

			QMenu::item {{
				background-color: transparent;
				spacing: 0px;
				border: 1px solid transparent;
				margin: 0px;
			}}
			''')

		self.setAttributes(**kwargs)


	@property
	def draggable_header(self):
		'''Get the draggable header.
		'''
		try:
			return self._draggable_header

		except AttributeError as error:
			from uitk.widgets.pushButtonDraggable import PushButtonDraggable
			dh = PushButtonDraggable()

			wAction = QtWidgets.QWidgetAction(self)
			wAction.setDefaultWidget(dh)
			self.insertAction_(wAction)

			self._draggable_header = dh
			return self._draggable_header


	@property
	def containsMenuItems(self):
		'''Query whether any child objects have been added to the menu.
		'''
		return bool(self.childWidgets)


	def addToOpenMenus(self):
		"""Adds the current instance of the menu to the list of open menus.
		"""
		self.openMenus.append(self)


	def removeFromOpenMenus(self):
		"""Removes the current instance of the menu from the list of open menus.
		If the instance is not in the list, no action is taken.
		"""
		try:
			self.openMenus.remove(self)
		except ValueError:
			pass


	@classmethod
	def hideAll(cls):
		"""Hides all currently open instances of the menu.
		"""
		for menu in cls.openMenus:
			menu.hide()


	@classmethod
	def hideLastOpened(cls):
		"""Hides the last opened instance of the menu, if one exists.
		If there are no open menus, no action is taken.
		"""
		try:
			cls.openMenus[-1].hide()
		except IndexError:
			pass


	def getChildWidgets(self, inc=[], exc=[]):
		'''Get a list of the menu's child widgets.

		:Parameters:
			inc (list): Include only widgets of the given type(s). ie. ['QCheckBox', 'QRadioButton']
			exc (list): Exclude widgets by type.
		:Return:
			(list) child widgets.
		'''
		if any((inc, exc)):
			return [w for w in self.childWidgets 
					if not w.__class__.__base__.__name__ in exc 
						and (w.__class__.__base__.__name__ in inc if inc 
					else w.__class__.__base__.__name__ not in inc)]
		else:
			return list(self.childWidgets)


	def setTitle(self, title=''):
		'''Set the menu's title to the given string.
		If no title is given, the fuction will attempt to use the menu parents text.

		Parameters:
			title (str): Text to apply to the menu's header.
		'''
		if not title:
			try:
				title = self.parent().text()
			except AttributeError as error:
				try:
					title = self.parent().currentText()
				except AttributeError as error:
					pass

		self.draggable_header.setText(title)
		super().setTitle(title)


	def getActionAtIndex(self, index):
		'''
		'''
		try:
			return self.actions()[1:][index] #slice the actions list to omit the header and any built-in hidden buttons (ie. 'apply', 'toggleAll').

		except IndexError as error:
			return None


	def insertAction_(self, wAction, index=-1):
		'''Extends insertAction to allow inserting by index.

		Parameters:
			wAction (obj): The widget action to insert.
			index (int): The desired index. (It appends the action if index is invalid)
		'''
		_wAction = self.getActionAtIndex(index)
		self.insertAction(_wAction, wAction) #insert before _wAction. It appends the action if before is nullptr or before is not a valid action for this widget


	def _addFormLayout(self, key='form', index=0):
		'''Create a two column form layout that can later be referenced using a key.

		Parameters:
			key (str)(int): The key identifier for the layout.
			index(int): The index corresponding to the vertical positioning of the layout.

		Return:
			(obj) QLayout.
		'''
		form = QtWidgets.QWidget(self)
		form.setStyleSheet(f'QWidget {{background-color:rgba(50,50,50,{self.alpha});}}')

		layout = QtWidgets.QFormLayout(form)
		layout.setVerticalSpacing(0)

		wAction = QtWidgets.QWidgetAction(self)
		wAction.setDefaultWidget(form)
		self.insertAction_(wAction, index)

		self.layouts[key] = layout
		return self.layouts[key]


	def getFormLayout(self, key='form'):
		'''Get a two column form layout using a key.

		Parameters:
			key (str)(int): The key identifier for the layout.

		Return:
			(obj) QLayout.
		'''
		try:
			return self.layouts[key]

		except (KeyError, AttributeError) as error: #create a layout at the given key.
			self.layouts[key] = self._addFormLayout(key)
			return self.layouts[key]


	def _addVBoxLayout(self, key='vBox', index=0):
		'''Create a single column vertical layout that can later be referenced using a key.

		Parameters:
			key (str)(int): The key identifier for the layout.
			index(int): The index corresponding to the vertical positioning of the layout.

		Return:
			(obj) QLayout.
		'''
		form = QtWidgets.QWidget(self)
		form.setStyleSheet(f'QWidget {{background-color:rgba(50,50,50,{self.alpha});}}')

		layout = QtWidgets.QVBoxLayout(form)
		layout.setContentsMargins(0,0,0,0)
		layout.setSpacing(0)

		wAction = QtWidgets.QWidgetAction(self)
		wAction.setDefaultWidget(form)
		self.insertAction_(wAction, index)

		self.layouts[key] = layout
		return self.layouts[key]


	def getVBoxLayout(self, key='vBox'):
		'''Get a vertical box layout using a key.

		Parameters:
			key (str)(int): The key identifier for the layout.

		Return:
			(obj) QLayout.
		'''
		try:
			return self.layouts[key]

		except (KeyError, AttributeError) as error: #create a layout at the given key.			
			self.layouts[key] = self._addVBoxLayout(key)
			return self.layouts[key]


	def _addApplyButton(self):
		'''Add a pushbutton that executes the parent object when pressed.
		The button is hidden by default.

		Return:
			(widget)
		'''
		if not self.parent():
			# print (f'# Error: {__file__} in _addApplyButton\n#\tOperation requires a parent widget.')
			return

		w = QtWidgets.QPushButton('Apply') #self.add('QPushButton', setText='Apply', setObjectName=self.parent().objectName(), setToolTip='Execute the command.')
		w.setObjectName('apply_button')
		w.setToolTip('Execute the command.')
		w.released.connect(lambda: self.parent().released.emit()) #trigger the released signal on the parent when the apply button is released.
		w.setMinimumSize(119, 26)

		layout = self.getVBoxLayout('menu_buttons') #get the 'menu_buttons' layout.
		layout.addWidget(w)
		w.hide()

		return w


	def _addToggleAllButton(self):
		'''Add a pushbutton that will uncheck any checkBoxes when pressed.
		The button is hidden by default.

		Return:
			(widget)
		'''
		w = QtWidgets.QPushButton('Uncheck All') #self.add('QPushButton', setText='Disable All', setObjectName='disableAll', setToolTip='Set all unchecked.')
		w.setObjectName('toggleAll')
		w.setToolTip('Toggle all checked|unchecked.')

		w.released.connect(lambda: [c.setChecked(not next(self.getChildWidgets(inc=['QCheckBox'])).checked()) for c in self.getChildWidgets(inc=['QCheckBox'])]) #trigger the released signal on the parent when the apply button is released.
		w.setMinimumSize(119, 26)

		layout = self.getVBoxLayout('menu_buttons') #get the 'menu_buttons' layout.
		layout.addWidget(w)
		w.hide()

		return w


	def add(self, widget, label='', checkableLabel=False, **kwargs):
		'''Add items to the QMenu.

		Parameters:
			widget (str)(obj): The widget to add. ie. 'QLabel', QtWidgets.QLabel, QtWidgets.QLabel()
			lable (str): Add a label. (which is actually a checkbox. by default it is not checkable)
			checkableLabel (bool): The label is checkable.

		additional kwargs:
			insertSeparator_ (bool): insert a separator before the widget.
			setLayoutDirection_ (str): ie. 'LeftToRight'
			setAlignment_ (str): ie. 'AlignVCenter'
			setButtonSymbols_ (str): ie. 'PlusMinus'
			setMinMax_ (str): Set the min, max, and step values with a string. ie. '1-100 step.1'

		Return:
			(obj) the added widget instance.

		ex.call:
		menu.add('QCheckBox', setText='Component Ring', setObjectName='chk000', setToolTip='Select component ring.')
		'''
		try: #get the widget
			w = getattr(QtWidgets, widget)() #ie. QtWidgets.QAction(self) object from string.
		except TypeError:
			w = widget() #if callable(widget) ie. QtWidgets.QAction(self) object.
		except:
			w = widget

		self.setAttributes(w, **kwargs) #set any additional given keyword args for the widget.

		type_ = w.__class__.__name__

		if type_=='QAction': #add action item.
			self.insertAction_(wAction)

		else:
			if self.menu_type=='form': #add widgets to the form layout.
				l = QtWidgets.QCheckBox()
				text = ''.join([(' '+i if i.isupper() else i) for i in label]).title() #format the attr name. ie. 'Subdivisions Height' from 'subdivisionsHeight'
				l.setText(text)
				l.setObjectName(label)
				if not checkableLabel:
					l.setCheckable(False)
					l.setStyleSheet('QCheckBox::hover {background-color: rgb(100,100,100); color: white;}')
				layout = self.getFormLayout() #get the default form layout.
				layout.addRow(l, w)
				self.childWidgets.add(l) #add the widget to the childWidgets list.

			else: #convert to action item, then add.
				layout = self.getVBoxLayout() #get the default vertical box layout.
				layout.addWidget(w)

			#set child height
			if w.sizeHint().width() > self.sizeHint().width():
				width = w.sizeHint().width() if w.sizeHint().width() >125 else 125
				w.setMinimumSize(width, w.sizeHint().height())
			try:
				l.setMinimumSize(l.sizeHint().width(), self.childHeight)
				# l.setMaximumSize(999, self.childHeight)
			except UnboundLocalError as error: #'l' does not exist. (not a form menu)
				pass

			self.childWidgets.add(w) #add the widget to the childWidgets list.

			setattr(self, w.objectName(), w) #add the widget's objectName as a QMenu attribute.

			self._addToContextMenuToolTip(w)

			if hasattr(w, 'released'): #Get the appropriate signal to connect to.
				w.released.connect(lambda w=w: self._setLastActiveChild(w)) #Connect the signal if one was found.
			elif hasattr(w, 'valueChanged'):
				w.valueChanged.connect(lambda value, w=w: self._setLastActiveChild(value, w))

		return w


	def _setLastActiveChild(self, widget, *args, **kwargs):
		'''Set the given widget as the last active.
		Maintains a list of the last 10 active child widgets.

		Parameters:
			widget = Widget to set as last active. The widget can later be returned by calling the 'lastActiveChild' method.
			*args **kwargs = Any additional arguments passed in by the wiget's signal during a connect call.

		Return:
			(obj) widget.
		'''
		# widget = args[-1]

		if not hasattr(self, '_lastActiveChild'):
			self._lastActiveChild = []

		del self._lastActiveChild[11:] #keep the list length at 10 elements.

		self._lastActiveChild.append(widget)
		# print(args, kwargs, widget.objectName() if hasattr(widget, 'objectName') else None)
		return self._lastActiveChild[-1]


	def lastActiveChild(self, name=False, asList=False):
		'''Get the given widget set as last active.
		Contains a list of the last 10 active child widgets.

		Parameters:
			name (bool): Return the last active widgets name as a string.

		Return:
			(obj)(str)(list) dependant on flags.

		ex. slot connection to the last active child widget:
			cmb.returnPressed.connect(lambda m=cmb.ctxMenu.lastActiveChild: getattr(self, m(name=1))()) #connect to the last pressed child widget's corresponding method after return pressed. ie. self.lbl000 if cmb.lbl000 was clicked last.
		'''
		if not hasattr(self, '_lastActiveChild'):
			return None

		if name:
			lastActive = str(self._lastActiveChild[-1].objectName())

		elif name and asList:
			lastActive = [str(w.objectName()) for w in self._lastActiveChild]

		elif asList:
			lastActive = [w for w in self._lastActiveChild]

		else:
			lastActive = self._lastActiveChild[-1]

		return lastActive


	def _addToContextMenuToolTip(self, menuItem):
		'''Add an item to the context menu toolTip.

		Parameters:
			menuItem (obj): The item to add.
		'''
		p = self.parent()
		if not all([self.menu_type=='context', p]):
			return

		if not hasattr(self, '_contextMenuToolTip'):
			self._contextMenuToolTip = '<u>Context menu items:</u>'
			p.setToolTip('{}<br><br>{}'.format(p.toolTip(), self._contextMenuToolTip)) #initialize the toolTip.

		try:
			contextMenuToolTip = '<b>{}</b> - {}'.format(menuItem.text(), menuItem.toolTip())
			p.setToolTip('{}<br>{}'.format(p.toolTip(), contextMenuToolTip))
		except AttributeError:
			pass


	def showEvent(self, event):
		'''
		:Parameters:
			event = <QEvent>
		'''
		self.resize(self.sizeHint().width(), self.sizeHint().height()+10) #self.setMinimumSize(width, self.sizeHint().height()+5)
		getCenter = lambda w, p: QtCore.QPoint(p.x()-(w.width()/2), p.y()-(w.height()/4)) #get widget center position.

		#set menu position
		if self.position=='cursorPos':
			pos = QtGui.QCursor.pos() #global position
			self.move(getCenter(self, pos)) #move to cursor position.

		elif isinstance(self.position, (tuple, list, set, QtCore.QPoint)): #position is given as a coordinate.
			if not isinstance(self.position, QtCore.QPoint):
				self.position = QtCore.QPoint(self.position[0], self.position[1])
			self.move(self.position)

		elif not isinstance(self.position, (type(None), str)): #if a widget is passed to 'position' (move to the widget's position).
			pos = getattr(self.positionRelativeTo.rect(), self.position)
			self.move(self.positionRelativeTo.mapToGlobal(pos()))

		elif self.parent(): #if parent: map relative to parent.
			pos = getattr(self.parent().rect(), self.position if not self.position=='cursorPos' else 'bottomLeft')
			pos = self.parent().mapToGlobal(pos())
			self.move(pos) # self.move(getCenter(self, pos))

			if self.getChildWidgets(inc=['QCheckBox']): #if the menu contains checkboxes:
				self.uncheckAllButton.show()

		super().showEvent(event)


	def setVisible(self, state):
		'''Called every time the widget is shown or hidden on screen.
		'''
		if state: #visible
			if not self.containsMenuItems: #prevent show if the menu is empty.
				return

			self.hideAll()
			self.addToOpenMenus()

			self.menu_timer.start(8000)  #5000 milliseconds = 5 seconds

			if not self.title():
				self.setTitle()

			if hasattr(self.parent(), 'released') and not self.parent().objectName()=='draggable_header':
				# print (f'show menu | title: {self.title()} | {self.parent().objectName()} has attr released.') #debug
				self.applyButton.show()

			checkboxes = self.getChildWidgets(inc=['QCheckBox'])
			if checkboxes: #returns None if the menu doesn't contain checkboxes.
				self.toggleAllButton.show()

		elif self.preventHide: #invisible
			return

		else:
			self.menu_timer.stop()
			self.removeFromOpenMenus()

		super().setVisible(state)


	def leaveEvent(self, event):
		'''
		Parameters:
			event = <QEvent>
		'''
		self.hide()

		super().leaveEvent(event)


	def hide(self, force=False):
		'''Sets the widget as invisible.
		Prevents hide event under certain circumstances.

		Parameters:
			force (bool): override preventHide.
		'''
		if force or not self.preventHide:

			for w in self.getChildWidgets():
				try:
					if w.view().isVisible(): #comboBox menu open.
						return
				except AttributeError as error:
					pass

			super().hide()



class MenuInstance():
	'''Get a Menu and ctxMenu instance.
	'''
	@property
	def menu_(self):
		'''Get the standard menu.
		'''
		try:
			return self._menu

		except AttributeError as error:
			self._menu = Menu(self, position='bottomLeft', menu_type='standard')
			return self._menu


	@property
	def ctxMenu(self):
		'''Get the context menu.
		'''
		try:
			return self._contextMenu

		except AttributeError as error:
			self._contextMenu = Menu(self, position='cursorPos', menu_type='context')
			return self._contextMenu

# -----------------------------------------------------------------------------









# -----------------------------------------------------------------------------

if __name__ == "__main__":
	import sys
	app = QtWidgets.QApplication.instance() or QtWidgets.QApplication(sys.argv) #return the existing QApplication object, or create a new one if none exists.

	# # create parent menu containing two submenus:
	# m = Menu(position='cursorPos')
	# m1 = Menu('Create', addMenu_=m)
	# m1.add('QAction', setText='Action', insertSeparator_=True)
	# m1.add('QAction', setText='Action', insertSeparator_=True)
	# m1.add('QPushButton', setText='Button')

	# m2 = Menu('Cameras', addMenu_=m)
	# m2.add('QAction', setText='Action', insertSeparator_=True)
	# m2.add('QAction', setText='Action', insertSeparator_=True)
	# m2.add('QPushButton', setText='Button')
	# m.show()

	# # form layout example
	menu = Menu(menu_type='form', padding=6, title='Title', position='cursorPos')
	s = menu.add('QDoubleSpinBox', label='attrOne', checkable=True, setSpinBoxByValue_=1.0)
	s = menu.add('QDoubleSpinBox', label='attrTwo', checkable=False, setSpinBoxByValue_=2.0)
	menu.show()

	print (menu.childWidgets)

	# m.exec_(parent=None)
	sys.exit(app.exec_())

# -----------------------------------------------------------------------------
# Notes
# -----------------------------------------------------------------------------

'''
Promoting a widget in designer to use a custom class:
>	In Qt Designer, select all the widgets you want to replace, 
		then right-click them and select 'Promote to...'. 

>	In the dialog:
		Base Class:		Class from which you inherit. ie. QWidget
		Promoted Class:	Name of the class. ie. "MyWidget"
		Header File:	Path of the file (changing the extension .py to .h)  ie. myfolder.mymodule.mywidget.h

>	Then click "Add", "Promote", 
		and you will see the class change from "QWidget" to "MyWidget" in the Object Inspector pane.
'''

# depricated:

	# def hide(self, force=False):
	# 	'''Sets the widget as invisible.
	# 	Prevents hide event under certain circumstances.

	# 	Parameters:
	# 		force (bool): override preventHide.
	# 	'''
	# 	if force or not self.preventHide:

	# 		for w in self.getChildWidgets():
	# 			try:
	# 				if w.view().isVisible(): #comboBox menu open.
	# 					return
	# 			except AttributeError as error:
	# 				pass

	# 		super().hide()


	# def show(self):
	# 	'''Show the menu.
	# 	'''
	# 	if not self.containsMenuItems: #prevent show if the menu is empty.
	# 		return

	# 	if not self.title():
	# 			self.setTitle()

	# 	if hasattr(self.parent(), 'released') and not self.parent().objectName()=='draggable_header':
	# 		# print (f'show menu | title: {self.title()} | {self.parent().objectName()} has attr released.') #debug
	# 		self.applyButton.show()

	# 	checkboxes = self.getChildWidgets(inc=['QCheckBox'])
	# 	if checkboxes: #returns None if the menu doesn't contain checkboxes.
	# 		self.toggleAllButton.show()

	# 	super().show()


	# def showEvent(self, event):
	# 	'''
	# 	Parameters:
	# 		event = <QEvent>
	# 	'''
	# 	self.resize(self.sizeHint().width(), self.sizeHint().height()+10) #self.setMinimumSize(width, self.sizeHint().height()+5)
	# 	getCenter = lambda w, p: QtCore.QPoint(p.x()-(w.width()/2), p.y()-(w.height()/4)) #get widget center position.

	# 	#set menu position
	# 	if self.position=='cursorPos':
	# 		pos = QtGui.QCursor.pos() #global position
	# 		self.move(getCenter(self, pos)) #move to cursor position.

	# 	elif not isinstance(self.position, (type(None), str)): #if a widget is passed to 'position' (move to the widget's position).
	# 		pos = getattr(self.positionRelativeTo.rect(), self.position)
	# 		self.move(self.positionRelativeTo.mapToGlobal(pos()))

	# 	elif self.parent(): #if parent: map relative to parent.
	# 		pos = getattr(self.parent().rect(), self.position if not self.position=='cursorPos' else 'bottomLeft')
	# 		pos = self.parent().mapToGlobal(pos())
	# 		self.move(pos) # self.move(getCenter(self, pos))

	# 	QtWidgets.QMenu.showEvent(self, event)